﻿using FliUtils.Mvvm;
using System;
using System.Collections.Generic;
using System.Windows;
using System.IO;
using System.Runtime.InteropServices;
using static FLIPilotTakeAPic.ViewModel.FliPro;

namespace FLIPilotTakeAPic.ViewModel
{
    public class MainWindowViewModel : ObservableProperties
    {
        //**************************************************************
        #region Private Constants

        // Invalid or closed device handle
        public const int INVALID_HANDLE = -1;

        public const string FILE_EXTENSION_RCD = ".rcd";

        public const int MaxImageWidth = 4096;

        #endregion Private Constants

        //**************************************************************
        #region Public Procedures

        public int Connect()
        {
            int status = -1;
            lock (DeviceLock)
            {
                if (!IsConnected)
                {
                    Handle = INVALID_HANDLE;
                    status = OpenDevice();

                    if (status < 0)
                    {
                        return status;
                    }

                    if (Handle != INVALID_HANDLE)
                    {
                        IsConnected = true;
                    }
                }
            }

            return status;
        }


        //-------------------------------------------
        /// <summary>
        /// Get a list of all connected cameras
        /// </summary>
        /// <returns>a list of Device Info structs representing all connected cameras</returns>
        public List<FPRODEVICEINFO> GetConnectedCameras()
        {
            FPRODEVICEINFO[] DevInfo = new FPRODEVICEINFO[FPRO_MAX_CONNECTED_CAMERAS];
            uint nCameras = FPRO_MAX_CONNECTED_CAMERAS;
            try
            {
                int value = NativeMethods.FPROCam_GetCameraList(DevInfo, ref nCameras);
            }
            catch (Exception ex)
            {
                // just return the empty list on any error
                nCameras = 0;
            }

            var devinfolist = new List<FPRODEVICEINFO>();   // start with an empty list
            for (int i = 0; i < nCameras; i++)
            {
                devinfolist.Add(DevInfo[i]);
            }

            return devinfolist;
        }

        public FliFrame GetCompletedFrame()
        {
            int bufsize = PixelsToBytes(ImageAreaWH.W, ImageAreaWH.H);
            FliFrame frame = new FliFrame((uint)bufsize);

            // Start the camera with framecount = 1
            SetStartCapture(1);   // Simply calls FPROFrame_CaptureStart(Handle, 1);

            // Calculate the Exposure and frame delay time, then call FPROFrame_GetVideoFrame(Handle, bufPtr, ref size, (uint)ExposureTimeInMS);
            // This will need to be called once for every framecount passed to SetStartCapture
            frame.RawDataLength = GetRawImageFrame(frame.RawData, bufsize, false);

            frame.IsValid = (frame.RawDataLength == bufsize);

            // Cleans up resources in the SDK when imaging is complete
            SetStopCapture();   // Simply calls FPROFrame_CaptureStop(Handle);

            return frame;
        }

        private long GetExposureTimeInMS(long value, ExposureUnits_Enum units)
        {
            double factor = 0;
            double realValue = value;
            switch (units)
            {
                case ExposureUnits_Enum.units_uSecs:
                    factor = 0.001;
                    break;
                case ExposureUnits_Enum.units_mSecs:
                    factor = 1;
                    break;
                case ExposureUnits_Enum.units_secs:
                    factor = 1000;
                    break;
                case ExposureUnits_Enum.units_mins:
                    factor = (long)1000 * 60;
                    break;
                case ExposureUnits_Enum.units_hours:
                    factor = (long)1000 * 3600;
                    break;
            }
            realValue *= factor;
            value = (int)realValue;

            return value;
        }

        public string GetValidFilename(string folder, string filename, string extension)
        {
            int fileIndex = 1;
            string newFileName = filename;
            DirectoryInfo di = new DirectoryInfo(folder);
            FileInfo[] TXTFiles = null;

            try
            {
                TXTFiles = di.GetFiles(filename + "*" + extension);

                List<int> indexList = new List<int>();
                for (int i = 0; i < TXTFiles.Length; i++)
                {
                    string tempName = TXTFiles[i].Name.Substring(filename.Length, TXTFiles[i].Name.Length - (filename.Length + 4)); // subtract off ".bin"
                    try
                    {
                        indexList.Add(Convert.ToInt32(tempName));
                    }
                    catch { }
                }
                while (indexList.Contains(fileIndex))
                {
                    fileIndex++;
                }
            }
            catch { }

            return folder + "\\" + newFileName + ((TXTFiles != null) && (TXTFiles.Length > 0) ? fileIndex.ToString() : String.Empty) + extension;
        }

        /// <summary>
        /// Starts the selected Grab operation using the Image Coordinates
        /// </summary>
        public void GrabClick()
        {
            if (!IsConnected)
            {
                ConnectToCamera();
            }

            if (IsConnected)
            { 
                // This function gets the image from the camera directly
                FliFrame frame = GetCompletedFrame();

                if ((frame != null) && frame.IsValid)
                {
                    // Only write valid frames here. Invalid frames are written in ParseRawImageData
                    WriteFileBasic(frame, RawFilePath);
                }
            }
        }

        // Creates the RCD file in the folder selected on the GUI
        public void SaveBINFile(FliFrame frame, string fileName)
        {
            try
            {
                FileStream fs = new FileStream(fileName, FileMode.CreateNew);
                // Create the writer for data.
                BinaryWriter w = new BinaryWriter(fs);

                if (frame.RawData == null)
                {
                    return;
                }
                else 
                {
                    w.Write(frame.RawData);
                }
                w.Close();
                fs.Close();
            }
            catch { }
        }

        public void WriteFileBasic(FliFrame frame, string folder)
        {
            if (!String.IsNullOrEmpty(folder))
            {
                try
                {
                    string fileName = String.Empty;
                    // Save the Combined file which is exactly how we received it from the board.
                    fileName = GetValidFilename(folder, "RCDData_", FILE_EXTENSION_RCD);
                    SaveBINFile(frame, fileName);
                }
                catch
                {
                }
            }
        }

        #endregion Public Procedures

        //**************************************************************
        #region Private Procedures

        private void ConnectToCamera()
        {
            List<FPRODEVICEINFO> connectedList = new List<FPRODEVICEINFO>();

            connectedList = GetConnectedCameras();

            if (connectedList.Count == 0)
            {
                MessageBox.Show("Could not find any attached cameras.", "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                return;
            }

            DeviceInfo = connectedList[0];

            int status = OpenDevice();

            if (Handle != INVALID_HANDLE)
            {
                IsConnected = true;

                // Pass the minimum commands to program the camera for Grab
                SetExposure(ExposureTime, ExposureDelayTime, ExposureUnits, ExposureDelayUnits, false);
                ImageAreaWH = new FliRectangleWH(0, 0, MaxImageWidth, MaxImageWidth);
                SetFrameType(FliFrameType.NORMAL);
            }
        }

        public long GetAbsoluteExposureTime(long value, ExposureUnits_Enum unit)
        {
            long factor = 1;

            switch (unit)
            {
                case ExposureUnits_Enum.units_uSecs:
                    if ((value < 20) && (value > 0))
                    {
                        value = 20;
                    }
                    factor = 1000;
                    break;
                case ExposureUnits_Enum.units_mSecs:
                    factor = 1000000;
                    break;
                case ExposureUnits_Enum.units_secs:
                    factor = 1000000000;
                    break;
                case ExposureUnits_Enum.units_mins:
                    factor = (long)1000000000 * 60;
                    break;
                case ExposureUnits_Enum.units_hours:
                    factor = (long)1000000000 * 3600;
                    break;
            }
            value *= factor;

            return value;
        }

        protected uint GetRawImageFrame(byte[] framebuf, int framesize, bool useDelay)
        {
            GCHandle bufGch = GCHandle.Alloc(framebuf, GCHandleType.Pinned);
            IntPtr bufPtr = bufGch.AddrOfPinnedObject();
            uint size = (uint)framesize;

            try
            {
                if (IsConnected)
                {
                    if (Handle < 0)
                    {
                        return size;
                    }

                    // The Exposure Time plus the Delay Time constitutes the entire wait time for FPROFrame_GetVideoFrame
                    long ExposureTimeInMS = GetExposureTimeInMS(ExposureTime, ExposureUnits);

                    int status = NativeMethods.FPROFrame_GetVideoFrame(Handle, bufPtr, ref size, (uint)ExposureTimeInMS);
                    if (status < 0)
                    {
                        MessageBox.Show("GetVideoFrame failed with a Semiphore error.", "Warning", MessageBoxButton.OK,
                            MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                    }
                }
            }
            finally
            {
                bufGch.Free();
            }
            return size;
        }

        protected int OpenDevice()
        {
            int status = -1;

            try
            {
                status = NativeMethods.FPROCam_Open(ref _DeviceInfo, out _Handle);
            }
            catch
            {
                status = -1;
            }
            return status;
        }

        private int PixelDepth = 12;     // 12 bits for each pixel
        private int MetaDataSize = 246;  // The size of the metadata coming from the camera

        public int PixelsToBytes(int width, int height)
        {
            height *= 2;   // Double the height because two images are captured

            int pixels = width * height;
            int bytes = MetaDataSize + (((pixels * (int)PixelDepth) + 7) / 8);
            return bytes;

        }

        protected void SetExposure(long value, long delay, ExposureUnits_Enum unit, ExposureUnits_Enum unitDelay, bool Immediate)
        {
            value = GetAbsoluteExposureTime(value, unit);
            delay = GetAbsoluteExposureTime(delay, unitDelay);
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    return;
                }
                NativeMethods.FPROCtrl_SetExposure(Handle, (long)value, (long)delay, Immediate);
            }
        }

        protected void SetFrameType(FliFrameType value)
        {
            NativeMethods.FPROFrame_SetFrameType(Handle, FPRO_FRAME_TYPE.FPRO_FRAMETYPE_NORMAL); // Set the Frame Type to Normal
        }

        protected void SetImageArea(FliRectangle value)
        {
            SetImageArea(new FliRectangleWH(value));
        }

        protected void SetImageArea(FliRectangleWH value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    return;
                }

                // NOTE: Camera always returns the entire width no matter what width is requested. 
                //       If a crop is desired it will have to be create in software afyter the image is captured.
                FliRectangleWH RequestedImageAreaWH = new FliRectangleWH(0, value.UL.Y, MaxImageWidth, value.H);
                int result = NativeMethods.FPROFrame_SetImageArea(Handle,
                   (uint)RequestedImageAreaWH.UL.X, (uint)RequestedImageAreaWH.UL.Y, (uint)RequestedImageAreaWH.W, (uint)RequestedImageAreaWH.H);
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Start a frame capture sequence
        /// </summary>
        /// <param name="frameCount">number of frames to capture. 0=no limit</param>
        /// <returns>false on any error</returns>
        protected bool SetStartCapture(int frameCount)
        {
            bool ok = false;
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    return ok;
                }

                // We only want to call FPROFrame_CaptureStart for Internal trigger
                int result = NativeMethods.FPROFrame_CaptureStart(Handle, (uint)frameCount);
                ok = (result >= 0);
            }
            return ok;
        }

        //-------------------------------------------
        /// <summary>
        /// Stop current frame capture sequence, if any)
        /// </summary>
        protected void SetStopCapture()
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    return;
                }
                NativeMethods.FPROFrame_CaptureStop(Handle);
            }
        }

        #endregion Private Procedures

        //**************************************************************
        #region Public Classes

        public class FliFrame
        {
            public FliFrame(uint rawDataLength)
            {
                RawData = new byte[rawDataLength];
                RawDataLength = rawDataLength;
                IsValid = false;
            }

            public byte[] RawData;
            public uint RawDataLength;
            public bool IsValid;
        }

        #endregion Public Classes

        //**************************************************************
        #region Observable Properties

        public FPRODEVICEINFO DeviceInfo
        {
            get { return _DeviceInfo; }
            set { SetValue<FPRODEVICEINFO>(ref _DeviceInfo, value); }
        }
        protected FPRODEVICEINFO _DeviceInfo;

        public int ExposureDelayTime
        {
            get { return _ExposureDelayTime; }
            set { SetValue<int>(ref _ExposureDelayTime, value); }
        }
        protected int _ExposureDelayTime = 0;

        public ExposureUnits_Enum ExposureDelayUnits
        {
            get { return _ExposureDelayUnits; }
            set { SetValue<ExposureUnits_Enum>(ref _ExposureDelayUnits, value); }
        }
        protected ExposureUnits_Enum _ExposureDelayUnits = ExposureUnits_Enum.units_mSecs;

        public int ExposureTime
        {
            get { return _ExposureTime; }
            set { SetValue<int>(ref _ExposureTime, value); }
        }
        protected int _ExposureTime = 50;

        public ExposureUnits_Enum ExposureUnits
        {
            get { return _ExposureUnits; }
            set { SetValue<ExposureUnits_Enum>(ref _ExposureUnits, value); }
        }
        protected ExposureUnits_Enum _ExposureUnits = ExposureUnits_Enum.units_mSecs;

        protected int Handle
        {
            get { return _Handle; }
            set { _Handle = value; }
        }
        protected int _Handle = INVALID_HANDLE;

        //-------------------------------------------
        /// <summary>
        /// Active image area
        /// </summary>
        public FliRectangle ImageArea
        {
            get { return _ImageArea; }
            set
            {
                if (SetValue<FliRectangle>(ref _ImageArea, value))
                {
                    SetImageArea(value);
                }
            }
        }
        protected FliRectangle _ImageArea;

        //-------------------------------------------
        /// <summary>
        /// Active Image area
        /// </summary>
        public FliRectangleWH ImageAreaWH
        {
            get { return new FliRectangleWH(_ImageArea); }
            set { ImageArea = new FliRectangle(value); }
        }

        //-------------------------------------------
        /// <summary>
        /// True if device is connected
        /// </summary>
        public bool IsConnected
        {
            get { return _IsConnected; }
            protected set { SetValue<bool>(ref _IsConnected, value); }
        }
        protected bool _IsConnected;

        public string RawFilePath
        {
            get { return _RawFilePath; }
            set { SetValue<string>(ref _RawFilePath, value); }
        }
        protected string _RawFilePath = String.Empty;

        #endregion

        //**************************************************************
        #region Protected Properties

        protected object DeviceLock;

        #endregion Public Events

        //**************************************************************
        #region Public Types

        public enum ExposureUnits_Enum
        {
            units_uSecs,
            units_mSecs,
            units_secs,
            units_mins,
            units_hours
        }

        #endregion Public Types

    }
}
