﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using static FliProCameraLib.FliPro;

namespace FliProCameraLib
{
    public class FliProFrameMetadata
    {

        //**************************************************************
        #region Constructor and Initialization

        public FliProFrameMetadata()
        {
            FliProCameraManager fliProCameraManager = FliProCameraManager.Instance;
            MagicNumber = MAGIC_NUMBER;
            MetadataLength = META_DATA_LENGTH_VERSION_2;
            MetadataVersion = META_DATA_VERSION_2;

            DateTime today = DateTime.UtcNow;
            CaptureYear = today.Year;
            CaptureMonth = today.Month;
            CaptureDay = today.Day;
            HorizontalBinning = 1;
            VerticalBinning = 1;
            FliPilotVersion = fliProCameraManager.GetFliPilotVersion();
            LibFliProVersion = fliProCameraManager.GetAPIVersion();
            string name = Assembly.GetExecutingAssembly().GetName().ToString();
            // Initialize the Unused portions of the Metadata header
            UnusedBytesAfterFPGATemperature = new byte[UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE];
            for (int i = 0; i < UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE; i++)
            {
                UnusedBytesAfterFPGATemperature[i] = 0xFF;
            }
            TemperatureSetPoint = 0;
            CoolerDutyCycle = 0;
            UnusedBytesAfterLongitude = new byte[UNUSED_BYTES_AFTER_LONGITUDE_SIZE];
            for (int i = 0; i < UNUSED_BYTES_AFTER_LONGITUDE_SIZE; i++)
            {
                UnusedBytesAfterLongitude[i] = 0xFF;
            }
            UnusedByteAfterDummyPixelsPerRow = 0xFF;
        }

        public FliProFrameMetadata(byte[] metadata)
        {
            int currentIndex = 0;
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];
            Regex regex = new Regex(@"[^a-zA-Z0-9\s\-]", (RegexOptions)0);

            if (metadata.Length < META_DATA_LENGTH_VERSION_1)
            {
                MagicNumber = 0;    // This will signal that the data is bad
                LowGain = 0;        // These are just to prevent an error
                HighGain = 0;
                return;
            }

            Array.Copy(metadata, currentIndex, fourBytes, 0, MAGIC_NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            MagicNumber = BitConverter.ToInt32(fourBytes, 0);
            currentIndex += MAGIC_NUMBER_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            MetadataLength = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += METADATA_LENGTH_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            MetadataVersion = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += METADATA_VERSION_SIZE;

            if (MetadataVersion == META_DATA_VERSION_1)
            {
                ParseMetadataVersion1(metadata, currentIndex, DateTime.UtcNow);
            }
            else
            {
                ParseMetadataVersion2(metadata, currentIndex, DateTime.UtcNow);
            }
        }

        public FliProFrameMetadata(byte[] metadata, DateTime captureDatetime)
        {
            int currentIndex = 0;
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];
            Regex regex = new Regex(@"[^a-zA-Z0-9\s\-]", (RegexOptions)0);

            if (metadata.Length < META_DATA_LENGTH_VERSION_1)
            {
                MagicNumber = 0;    // This will signal that the data is bad
                LowGain = 0;        // These are just to prevent an error
                HighGain = 0;
                return;
            }

            Array.Copy(metadata, currentIndex, fourBytes, 0, MAGIC_NUMBER_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            MagicNumber = BitConverter.ToInt32(fourBytes, 0);
            currentIndex += MAGIC_NUMBER_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            MetadataLength = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += METADATA_LENGTH_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            MetadataVersion = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += METADATA_VERSION_SIZE;

            if (MetadataVersion == META_DATA_VERSION_1)
            {
                ParseMetadataVersion1(metadata, currentIndex, captureDatetime);
            }
            else
            {
                ParseMetadataVersion2(metadata, currentIndex, captureDatetime);
            }
        }

        private byte CheckForSignMagnitude(byte value, ref int signed)
        {
            if ((value & NEGATIVE_BYTE_MASK) == NEGATIVE_BYTE_MASK)
            {
                signed = -1;
                value = (byte)(value & NEGATIVE_BYTE_CLEAR_MASK);
            }
            else
            {
                signed = 1;
            }
            return value;
        }

        #endregion Constructor and Initialization

        //**************************************************************
        #region Offset Constants

        private const int MAGIC_NUMBER_SIZE = 4;                            // 4	Bytes					x4D657461	
        private const int METADATA_LENGTH_SIZE = 2;                         // 2 Bytes
        private const int METADATA_VERSION_SIZE = 2;                        // 2 Bytes
        private const int CAMERA_MODEL_SIZE = 40;                           // 40 Char
        private const int CAMERA_SERIAL_NUMBER_SIZE = 24;                   // 24 Char
        private const int CAPTURE_YEAR_SIZE = 1;                            // 1	Bytes                   Value = (7:1) Year-2000 , (0) Month MSb			
        private const int CAPTURE_MONTH_SIZE = 1;                           // 1	Bytes                   Value = (7:5) Month, (4:0) Day	
        //private const int CaptureDay { get; set; }
        private const int CAPTURE_TIME_SIZE = 5;                            // 5	Bytes                   Value = (39:0) Time in 100ns periods				ISO-8601
        private const int SENSOR_PIXEL_BIT_DEPTH_SIZE = 1;                  // 1	Bytes                   Value = Bit Depth	
        private const int DATA_PIXEL_BIT_DEPTH_ZERO_POINT_SIZE = 1;         // 1	Bytes                   Value = (Bit Depth - 1) (3:0),  ZeroPoint(7:4)	
        private const int HORIZONTAL_PIXELS_SIZE = 2;                       // 2	Bytes
        private const int VERTICAL_PIXELS_SIZE = 2;                         // 2	Bytes
        private const int EXPOSURE_TIME_SIZE = 4;                           // 4	Bytes                   Value = Exposure Time in 10µs periods	
        private const int SENSOR_BASE_TEMP_SIZE = 2;                        // 2	Bytes
        private const int SENSOR_TEMP_SIZE = 2;                             // 2	Bytes
        private const int GLOBAL_GAIN_SIZE = 1;                             // 1	Bytes
        private const int LDR_GAIN_SIZE = 1;                                // 1	Bytes                   Value = (2:0) for 400 Sensor	
        private const int HDR_GAIN_SIZE = 1;                                // 1	Bytes                   Value = (2:0) for 400 Sensor		
        private const int DUMMY_PIXELS_PER_ROW_SIZE = 1;                    // 1	Bytes
        private const int UNUSED_BYTE_AFTER_DUMMY_PIXELS_PER_ROW_SIZE = 1;  // 1	Bytes
        private const int CONTROL_BLOCK_SIZE = 1;                           // 1	Bytes

        private const int DEAD_PIXEL_CORRECTION_MASK = 0x80;                // 7								
        private const int PIXEL_ORDERED_IMAGE_MASK = 0x40;                  // 6								
        private const int HDR_MODE = 0x20;                                  // 5								
        private const int LOW_NOISE_MASK = 0x10;                            // 4								
        private const int LOW_DARK_CURRENT_MASK = 0x08;                     // 3								
        private const int HORIZONTAL_SCAN_DIRECTION_INVERT_MASK = 0x04;     // 2								
        private const int VERTICAL_SCAN_DIRECTION_INVERT_MASK = 0x02;       // 1	
        private const int NON_ROW_ALLIGNED_IMAGE_MASK = 0x01;               // 0	

        private const int HORIZONTAL_BINNING_SIZE = 1;                      // 1	Bytes									
        private const int VERTICAL_BINNING_SIZE = 1;                        // 1	Bytes									
        private const int BLACK_LEVEL_ADJUST_SIZE = 3;                      // 2	Bytes				    Value = (23:0) for 400 Sensor				
        private const int BLACK_LEVEL_SUN_SIZE = 2;                         // 2	Bytes					Value = (5:0) for 400 Sensor				
        private const int SHUTTER_OPEN_DELAY_SIZE = 2;                      // 2	Bytes					Value = Signed				
        private const int SHUTTER_CLOSE_DELAY_SIZE = 2;                     // 2	Bytes					Value = Signed				
        private const int ILLUMINATION_START_DELAY_SIZE = 2;                // 2	Bytes					Value = Signed				
        private const int ILLUMINATION_STOP_DELAY_SIZE = 2;                 // 2	Bytes					Value = Signed				
        private const int TRACKING_START_ROW_SIZE = 2;                      // 4	Bytes					Value = (31:16) Start Row, (15:0) Start Column		
	    private const int TRACKING_START_COLUMN_SIZE = 2;	
        private const int TRACKING_STOP_ROW_SIZE = 2;                       // 4	Bytes					Value = (31:16) Stop Row, (15:0) Stop Column	
	    private const int TRACKING_STOP_COLUMN_SIZE = 2;		
        private const int IMAGE_START_ROW_SIZE = 2;                         // 4	Bytes					Value = (31:16) Start Row, (15:0) Start Column	
	    private const int IMAGE_START_COLUMN_SIZE = 2;		
        private const int IMAGE_STOP_ROW_SIZE = 2;                          // 4	Bytes					Value = (31:16) Stop Row, (15:0) Stop Column
	    private const int IMAGE_STOP_COLUMN_SIZE = 2;			
        private const int TRACKING_FRAMES_PER_IMAGE_FRAME_SIZE = 2;         // 2	Bytes									
        private const int FRAME_NUMBER_SIZE = 2;                    	    // 2	Bytes
        private const int STARTING_EXPOSURE_ROW_SIZE = 2;              	    // 2	Bytes
        private const int PRE_DUMMY_ROWS_SIZE = 3;                          // 3	Bytes                   Value = (23:12) Pre-Rows (12 bits), (11:0) Post-Rows (12 bits)		
        private const int NUMBER_OF_CHANNELS_SIZE = 1;              	    // 1	Bytes
        private const int CONTROL_BLOCK_2_SIZE = 1;              	        // 1	Bytes

        private const int BACK_SIDE_ILLUMINATED_MASK = 0x80;                // 7
        private const int BACK_SIDE_ILLUMINATED_ON_MASK = 0xFF;             // 7
        private const int BACK_SIDE_ILLUMINATED_OFF_MASK = 0x7F;            // 7
        private const int GPS_LOCKED_MASK = 0x40;                           // 6
        private const int GPS_LOCKED_ON_MASK = 0xFF;                        // 6
        private const int GPS_LOCKED_OFF_MASK = 0xBF;                       // 6

        private const int BASE_TEMPERATURE_SIZE = 2;              	        // 2	Bytes
        private const int FPGA_TEMPERATURE_SIZE = 2;              	        // 2	Bytes

        private const int UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE = 2;     // 2	Bytes

        private const int CONTROL_BLOCK_TEST_SIZE = 1;              	    // 1	Bytes

        // Version 2 Masks
        public const uint DIRECTION_FLAG_MASK = 0x80000000;
        private const uint DIRECTION_DEGREES_MASK = 0x7FFFFFFF;
        public const int DIRECTION_DEGREES_DIVISOR = 600000;

        // The Firmware doesn't know about these masks so 1 means off
        private const int TEST_USE_SHIFTED_AVERAGING_MASK = 0x80;           // 7
        private const int TEST_USE_SHIFTED_AVERAGING_ON_MASK = 0x7F;        // 7
        private const int TEST_USE_SHIFTED_AVERAGING_OFF_MASK = 0xFF;       // 7
        private const int TEST_IS_STACKED_FRAME_MASK = 0x40;                // 6
        private const int TEST_IS_HIGH_FRAME_MASK = 0x20;                   // 5
        private const int TEST_IS_MERGED_FRAME_MASK = 0x10;                 // 4

        private const int TEST_LOW_GAIN_ABSOLUTE_SIZE = 2;                  // 2	Bytes
        private const int TEST_HIGH_GAIN_ABSOLUTE_SIZE = 2;                 // 2	Bytes

        private const int FIRMWARE_VERSION_SIZE = 4;                        // 4	Bytes

        // New for Version 2
        private const int TEMPERATURE_SET_POINT_SIZE = 2;                   // 2	Bytes
        private const int COOLER_DUTY_CYCLE_SIZE = 1;                       // 1	Bytes
        private const int TIMESTAMP_SIZE = 30;                              // 30   Bytes                   ISO8601 Timestamp  yyyy-mm-ddThh:mm:ss.ffffff
        private const int LATITUDE_SIZE = 4;                                // 4	Bytes
        private const int LONGITUDE_SIZE = 4;                               // 4	Bytes
        private const int FLIPILOT_VERSION_SIZE = 12;                       // 12	Bytes
        private const int LIB_FLI_PRO_VERSION_SIZE = 12;                    // 12	Bytes
        private const int HORIZONTAL_PIXEL_SIZE = 1;                        // 1	Bytes
        private const int VERTICAL_PIXEL_SIZE = 1;                          // 1	Bytes
        private const int UNUSED_BYTES_AFTER_LONGITUDE_SIZE = 30;           // 30	Bytes

        public const int MAGIC_NUMBER = 0x4D657461;
        public const short META_DATA_LENGTH_VERSION_1 = 156;
        public const short META_DATA_LENGTH_VERSION_2 = 246;
        public const short META_DATA_VERSION_1 = 1;
        public const short META_DATA_VERSION_2 = 2;
        public const byte NEGATIVE_BYTE_MASK = 0x80;
        public const byte NEGATIVE_BYTE_CLEAR_MASK = 0x7F;
        public const ushort NEGATIVE_SHORT_MASK = 0x8000;
        public const ushort NEGATIVE_SHORT_CLEAR_MASK = 0x7FFF;

        #endregion Offset Constants

        //**************************************************************
        #region Public Constants

        public const int MILLISECS_PER_HOUR = 3600 * 1000;
        public const int MILLISECS_PER_MIN = 60 * 1000;
        public const int MILLISECS_PER_SEC = 1000;

        #endregion Public Constants

        //**************************************************************
        #region Public Properties

        // Starting Byte
        public long MagicNumber { get; set; }                       // 4	Bytes			0		x4D657461	
        public int MetadataLength { get; set; }                     // 2    Bytes           4
        public int MetadataVersion { get; set; }                    // 2    Bytes           6
        public string CameraModel { get; set; }                     // 40   Char            8
        public string CameraSerialNumber { get; set; }              // 24   Char           48
        public string CaptureString = String.Empty;                 // 30   Bytes                   ISO8601 Timestamp  yyyy-mm-ddThh:mm:ss.ffffff
        public int CaptureYear { get; set; }                        // 1	Bytes          72       Value = (7:1) Year-2000 , (0) Month MSb			
        public int CaptureMonth { get; set; }                       // 1	Bytes          73       Value = (7:5) Month, (4:0) Day	
        public int CaptureDay { get; set; }
        public int CaptureHour { get; set; }
        public int CaptureMinute { get; set; }
        public int CaptureSecond { get; set; }
        public int CaptureNanoseconds { get; set; }
        public long CaptureTime { get; set; }                       // 5	Bytes          74       Value = (39:0) Time in 100ns periods				ISO-8601
        public int SensorPixelBitDepth { get; set; }                // 1	Bytes          79       Value = Bit Depth	
        public int DataPixelBitDepth { get; set; }                  // 1	Bytes          80       Value = (Bit Depth - 1) (3:0),  ZeroPoint(7:4)	
        public int DataZeroPoint { get; set; }     	
        public int HorizontalPixels { get; set; }                   // 2	Bytes          81
        public int VerticalPixels { get; set; }                     // 2	Bytes          83
        public int ExposureTime { get; set; }                       // 4	Bytes          85       Value = Exposure Time in 10us periods	
        public int SensorColdTemp { get; set; }                     // 2	Bytes          89       Value = signed 1/10 degree C
        public int SensorTemp { get; set; }                         // 2	Bytes          91       Value = signed 1/10 degree C
        public int GlobalGain { get; set; }                         // 1	Bytes          93
        public int LowGain { get; set; }                            // 1	Bytes          94       Value = (2:0) for 400 Sensor	
        public int HighGain { get; set; }                            // 1	Bytes          95       Value = (2:0) for 400 Sensor		
        public int DummyPixelsPerRow { get; set; }                  // 1	Bytes          96
        public int UnusedByteAfterDummyPixelsPerRow { get; set; }   // 1	Bytes          97       		
        public int ControlBlock { get; set; }                       // 1	Bytes          98

        public bool DeadPixelCorrection { get; set; }               // 7				   			
        public bool PixelOrderedImage { get; set; }                 // 6						   1 means all the pixels for each row are in order	
        public bool HDRMode { get; set; }                           // 5								
        public bool LowNoise { get; set; }                          // 4								
        public bool LowDarkCurrent { get; set; }                    // 3								
        public bool HorizontalScanDirectionInvert { get; set; }     // 2								
        public bool VerticalScanDirectionInvert { get; set; }       // 1	
        public bool NonRowAllignedImage { get; set; }               // 0	                       1 means that row 0 starts at StartingExposureRow

        public int HorizontalBinning { get; set; }                  // 1	Bytes		   99						
        public int VerticalBinning { get; set; }                    // 1	Bytes	      100						
        public int BlackLevelAdjust { get; set; }                   // 3	Bytes		  101	    Value = (23:0) for 400 Sensor				
        public int BlackLevelSun { get; set; }                      // 2	Bytes		  104		Value = (5:0) for 400 Sensor				
        public int ShutterOpenDelay { get; set; }                   // 2	Bytes		  106		Value = Signed				
        public int ShutterCloseDelay { get; set; }                  // 2	Bytes		  108		Value = Signed				
        public int IlluminationStartDelay { get; set; }             // 2	Bytes		  110		Value = Signed				
        public int IlluminationStopDelay { get; set; }              // 2	Bytes		  112		Value = Signed				
        public int TrackingStartRow { get; set; }                   // 4	Bytes		  114		Value = (31:16) Start Row, (15:0) Start Column
	    public int TrackingStartColumn { get; set; }			
        public int TrackingStopRow { get; set; }                    // 4	Bytes		  118		Value = (31:16) Stop Row, (15:0) Stop Column
	    public int TrackingStopColumn { get; set; }			
        public int ImageStartRow { get; set; }                      // 4	Bytes		  122		Value = (31:16) Start Row, (15:0) Start Column	
	    public int ImageStartColumn { get; set; }		
        public int ImageStopRow { get; set; }                       // 4	Bytes		  126		Value = (31:16) Stop Row, (15:0) Stop Column
	    public int ImageStopColumn { get; set; }			
        public int TrackingFramesPerImageFrame { get; set; }        // 2	Bytes		  130						
        public int FrameNumber { get; set; }                    	// 2	Bytes         132
        public int StartingExposureRow { get; set; }                // 2	Bytes         134
        public int PreDummyRows { get; set; }                       // 3	Bytes         136       Value = (23:12) Pre-Rows (12 bits), (11:0) Post-Rows (12 bits)		
        public int PostDummyRows { get; set; }
        public int NumberOfDataChannels { get; set; }               // 1	Bytes		  139	
        public int ControlBlock2 { get; set; }                      // 1	Bytes         140	

        public bool BackSideIlluminated { get; set; }               // 7							1 means back side illuminated			   			
        public bool GPSLocked { get; set; }                         // 6							1 means the GPS unit is locked onto the satellite			   			

        public int BaseTemperature { get; set; }                    // 2	Bytes         141       Tenths of a degree C starting at 0
        public int FPGATemperature { get; set; }                    // 2	Bytes         143       Tenths of a degree C starting at 0

        public byte[] UnusedBytesAfterFPGATemperature { get; set; } // 2	Bytes         145

        public int ControlBlockTest { get; set; }                   // 1	Bytes         147	

        public bool UseShiftedAveraging { get; set; }               // 7							0 means each pixel was multiplied by 16	
        public bool IsStackedFrame { get; set; }                    // 6							0 means this frame was generated by stacking "FrameNumber" images	
        public bool IsHighFrame { get; set; }                       // 5							0 means this frame is a High image	
        public bool IsMergedFrame { get; set; }                     // 4							0 means this frame is a Merged image	

        public int TestLowGainAbsolute { get; set; }                // 2	Bytes         148
        public int TestHighGainAbsolute { get; set; }               // 2	Bytes         150

        public uint FirmwareVersion { get; set; }                   // 4	Bytes		  152	

        public int TemperatureSetPoint { get; set; }                // 2	Bytes         156
        public byte CoolerDutyCycle { get; set; }                   // 1	Bytes         157

        public bool LatitudeNorth = true;
        public int LatitudeDegrees = 0;
        public double LatitudeMinutes = 0;
        public uint LatitudeRaw = 0;

        public bool LongitudeEast = true;
        public int LongitudeDegrees = 0;
        public double LongitudeMinutes = 0;
        public uint LongitudeRaw = 0;

        public string FliPilotVersion { get; set; }                 // 12   Char          190
        public string LibFliProVersion { get; set; }                // 12   Char          202
        public byte HorizontalPixelSize { get; set; }               //  1	Byte          214     (measured in 0.1 µm)
        public byte VerticalPixelSize { get; set; }                 //  1	Byte          215     (measured in 0.1 µm)
        public byte[] UnusedBytesAfterLongitude { get; set; }       // 30	Bytes         216
           
        #endregion Public Properties

        //**************************************************************
        #region Public Functions

        public override string ToString()
        {
            FliProCameraManager fliProCameraManager = FliProCameraManager.Instance;
            string value = String.Empty;

            if (MetadataVersion == META_DATA_VERSION_1)
            {
                string unusedArray = String.Empty;

                for (int i = 0; i < UnusedBytesAfterFPGATemperature.Length; i++)
                {
                    unusedArray += (i == 0 ? String.Empty : ", ") + UnusedBytesAfterFPGATemperature[i].ToString("X2");
                }

                value = "MagicNumber           = 0x" + MagicNumber.ToString("X8") + Environment.NewLine +                                                      // 4	Bytes					x4D657461
                        "MetadataLength        = " + MetadataLength.ToString() + Environment.NewLine +                                                         // 2    Bytes
                        "MetadataVersion       = " + MetadataVersion.ToString() + Environment.NewLine +                                                        // 2    Bytes
                        "CameraModel           = " + CameraModel.ToString() + Environment.NewLine +                                                            // 40   Char
                        "CameraSerialNumber    = " + CameraSerialNumber.ToString() + Environment.NewLine +                                                     // 24   Char
                        "CaptureYear           = " + CaptureYear.ToString() + Environment.NewLine +                                                            // 1	Bytes                   Value = (7:1) Year-2000 , (0) Month MSb			
                        "CaptureMonth          = " + CaptureMonth.ToString() + Environment.NewLine +                                                           // 1	Bytes                   Value = (7:5) Month, (4:0) Day	
                        "CaptureDay            = " + CaptureDay.ToString() + Environment.NewLine +
                        "CaptureTime           = " + FliProCamera.GetClockFromTime(CaptureTime / 10000, FliProCamera.ExposureUnits_Enum.units_mSecs) + " ("
                                                   + CaptureTime.ToString() + " (100ns periods))" + Environment.NewLine +                                      // 5	Bytes                   Value = (39:0) Time in 100ns periods				ISO-8601
                        "SensorPixelBitDepth   = " + SensorPixelBitDepth.ToString() + Environment.NewLine +                                                    // 1	Bytes                   Value = Bit Depth	
                        "DataPixelBitDepth     = " + DataPixelBitDepth.ToString() + " (value stored as value-1)" + Environment.NewLine +                       // 1	Bytes                   Value = (Bit Depth - 1) (3:0),  ZeroPoint(7:4)	
                        "DataZeroPoint         = " + DataZeroPoint.ToString() + Environment.NewLine +
                        "HorizontalPixels      = " + HorizontalPixels.ToString() + " (0x" + HorizontalPixels.ToString("X4") + ")" + Environment.NewLine +      // 2	Bytes
                        "VerticalPixels        = " + VerticalPixels.ToString() + " (0x" + VerticalPixels.ToString("X4") + ")" + Environment.NewLine +          // 2	Bytes
                        "ExposureTime          = " + (fliProCameraManager.DisplayCamera.DeviceType == (int)FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400 ?
                                                      FliProCamera.GetClockFromTime((long)((HDRMode ? (ExposureTime * 2 * 10.26) : ExposureTime * 10.26) / 1000),
                                                       FliProCamera.ExposureUnits_Enum.units_mSecs) + " (" + ExposureTime.ToString() +
                                                       (HDRMode ? " (20.52 µs periods))" : " (10.26 µs periods))") :
                                                       FliProCamera.GetClockFromTime((long)(ExposureTime * 10.32 / 1000), FliProCamera.ExposureUnits_Enum.units_mSecs) +
                                                       " (" + ExposureTime.ToString() + " (10.32 µs periods))") + Environment.NewLine +                               // 4	Bytes                   Value = Exposure Time in 10us periods	
                        "SensorColdTemp        = " + ((double)SensorColdTemp / 10).ToString("0.0") + " (0x" + ((short)SensorColdTemp).ToString("X4") + ")" + Environment.NewLine +             // 2	Bytes
                        "SensorTemp            = " + GetConvertedTemperature(SensorTemp) + " (0x" + ((short)SensorTemp).ToString("X4") + ")" + Environment.NewLine +                  // 2	Bytes
                        "GlobalGain            = " + GlobalGain.ToString() + " (0x" + GlobalGain.ToString("X2") + ")" + Environment.NewLine +                  // 1	Bytes
                        "LowGain               = " + LowGain.ToString() + " (0x" + LowGain.ToString("X2") + ")" + Environment.NewLine +                        // 1	Bytes                   Value = (2:0) for 400 Sensor	
                        "HighGain               = " + HighGain.ToString() + " (0x" + HighGain.ToString("X2") + ")" + Environment.NewLine +                        // 1	Bytes                   Value = (2:0) for 400 Sensor		
                        "ReferencePixelsPerRow = " + DummyPixelsPerRow.ToString() + " (0x" + DummyPixelsPerRow.ToString("X2") + ")" + Environment.NewLine +    // 1	Bytes
                        "UnusedByte            = " + UnusedByteAfterDummyPixelsPerRow.ToString() + " (0x" + UnusedByteAfterDummyPixelsPerRow.ToString("X2") + ")" + Environment.NewLine +                  // 1	Bytes
                        "ControlBlock          = " + ControlBlock.ToString() + " (0x" + ControlBlock.ToString("X2") + ")" + Environment.NewLine + Environment.NewLine +            // 1	Bytes

                        "DeadPixelCorrection   = " + DeadPixelCorrection.ToString() + Environment.NewLine +                                  // 7								
                        "PixelOrderedImage     = " + PixelOrderedImage.ToString() + Environment.NewLine +                                    // 6								
                        "HDRMode               = " + HDRMode.ToString() + Environment.NewLine +                                              // 5								
                        "LowNoise              = " + LowNoise.ToString() + Environment.NewLine +                                             // 4								
                        "LowDarkCurrent        = " + LowDarkCurrent.ToString() + Environment.NewLine +                                       // 3								
                        "HorzScanDirInvert     = " + HorizontalScanDirectionInvert.ToString() + Environment.NewLine +                        // 2								
                        "VertScanDirInvert     = " + VerticalScanDirectionInvert.ToString() + Environment.NewLine +                          // 1	
                        "NonRowAllignedImage   = " + NonRowAllignedImage.ToString() + Environment.NewLine + Environment.NewLine +            // 0	

                        "HorizontalBinning     = " + HorizontalBinning.ToString() + " (0x" + HorizontalBinning.ToString("X2") + ")" + Environment.NewLine +                        // 1	Bytes									
                        "VerticalBinning       = " + VerticalBinning.ToString() + " (0x" + VerticalBinning.ToString("X2") + ")" + Environment.NewLine +                            // 1	Bytes									
                        "BlackLevelAdjust      = " + BlackLevelAdjust.ToString() + " (0x" + BlackLevelAdjust.ToString("X4") + ")" + Environment.NewLine +                          // 3	Bytes				    Value = (13:0) for 400 Sensor				
                        "BlackLevelSun         = " + BlackLevelSun.ToString() + " (0x" + BlackLevelSun.ToString("X4") + ")" + Environment.NewLine +                                // 2	Bytes					Value = (5:0) for 400 Sensor				
                        "ShutterOpenDelay      = " + ShutterOpenDelay.ToString() + " (0x" + ((short)ShutterOpenDelay).ToString("X4") + ")" + Environment.NewLine +                 // 2	Bytes					Value = Signed				
                        "ShutterCloseDelay     = " + ShutterCloseDelay.ToString() + " (0x" + ((short)ShutterCloseDelay).ToString("X4") + ")" + Environment.NewLine +               // 2	Bytes					Value = Signed				
                        "IlluminStartDelay     = " + IlluminationStartDelay.ToString() + " (0x" + ((short)IlluminationStartDelay).ToString("X4") + ")" + Environment.NewLine +     // 2	Bytes					Value = Signed				
                        "IlluminStopDelay      = " + IlluminationStopDelay.ToString() + " (0x" + ((short)IlluminationStopDelay).ToString("X4") + ")" + Environment.NewLine +       // 2	Bytes					Value = Signed				
                        "TrackingStartRow      = " + TrackingStartRow.ToString() + " (0x" + TrackingStartRow.ToString("X4") + ")" + Environment.NewLine +                          // 4	Bytes					Value = (31:16) Start Row, (15:0) Start Column
                        "TrackingStartColumn   = " + TrackingStartColumn.ToString() + " (0x" + TrackingStartColumn.ToString("X4") + ")" + Environment.NewLine +
                        "TrackingStopRow       = " + TrackingStopRow.ToString() + " (0x" + TrackingStopRow.ToString("X4") + ")" + Environment.NewLine +                            // 4	Bytes					Value = (31:16) Stop Row, (15:0) Stop Column
                        "TrackingStopColumn    = " + TrackingStopColumn.ToString() + " (0x" + TrackingStopColumn.ToString("X4") + ")" + Environment.NewLine +
                        "ImageStartRow         = " + ImageStartRow.ToString() + " (0x" + ImageStartRow.ToString("X4") + ")" + Environment.NewLine +                                // 4	Bytes					Value = (31:16) Start Row, (15:0) Start Column	
                        "ImageStartColumn      = " + ImageStartColumn.ToString() + " (0x" + ImageStartColumn.ToString("X4") + ")" + Environment.NewLine +
                        "ImageStopRow          = " + ImageStopRow.ToString() + " (0x" + ImageStopRow.ToString("X4") + ")" + Environment.NewLine +                                  // 4	Bytes					Value = (31:16) Stop Row, (15:0) Stop Column
                        "ImageStopColumn       = " + ImageStopColumn.ToString() + " (0x" + ImageStopColumn.ToString("X4") + ")" + Environment.NewLine +
                        "TrackingFramesPerImageFrame = " + TrackingFramesPerImageFrame.ToString() + " (0x" + TrackingFramesPerImageFrame.ToString("X4") + ")" + Environment.NewLine +     // 2	Bytes									
                        "FrameNumber           = " + FrameNumber.ToString() + " (0x" + FrameNumber.ToString("X4") + ")" + Environment.NewLine +                                    // 2	Bytes
                        "StartingExposureRow   = " + StartingExposureRow.ToString() + " (0x" + StartingExposureRow.ToString("X4") + ")" + Environment.NewLine +                    // 2	Bytes
                        "PreReferenceRows      = " + PreDummyRows.ToString() + Environment.NewLine +                                                                               // 3	Bytes                   Value = (23:12) Pre-Rows (12 bits), (11:0) Post-Rows (12 bits)		
                        "PostReferenceRows     = " + PostDummyRows.ToString() + Environment.NewLine +
                        "NumberOfDataChannels  = " + NumberOfDataChannels.ToString() + Environment.NewLine +                                                                       // 1	Bytes
                        "ControlBlock2         = " + ControlBlock2.ToString() + " (0x" + ControlBlock2.ToString("X2") + ")" + Environment.NewLine + Environment.NewLine +          // 1	Bytes

                        "BackSideIlluminated   = " + BackSideIlluminated.ToString() + Environment.NewLine + Environment.NewLine +                                                  // 7

                        "BaseTemperature       = " + BaseTemperature.ToString() + Environment.NewLine +                                                                            // 1	Bytes
                        "FPGATemperature       = " + FPGATemperature.ToString() + Environment.NewLine +                                                                            // 1	Bytes

                        "UnusedBytes           = [" + unusedArray + "]" + Environment.NewLine + Environment.NewLine +                                                              // 2

                        "ControlBlockTest      = " + ControlBlockTest.ToString() + " (0x" + ControlBlockTest.ToString("X2") + ")" + Environment.NewLine + Environment.NewLine +    // 1	Bytes

                        "TestUseShiftedAveraging = " + UseShiftedAveraging.ToString() + Environment.NewLine +                                            // 7
                        "TestIsStackedFrame    = " + IsStackedFrame.ToString() + Environment.NewLine +                                                   // 6
                        "TestIsHighFrame       = " + IsHighFrame.ToString() + Environment.NewLine +                                                      // 5
                        "TestIsMergedFrame     = " + IsMergedFrame.ToString() + Environment.NewLine + Environment.NewLine +                              // 4

                        "TestLowGainAbsolute   = " + GetGainAbsoluteString(TestLowGainAbsolute) + Environment.NewLine +                                  // 2	Bytes
                        "TestHighGainAbsolute  = " + GetGainAbsoluteString(TestHighGainAbsolute) + Environment.NewLine + Environment.NewLine +           // 2	Bytes

                        "FirmwareVersion       = " + "0x" + FirmwareVersion.ToString("X8") + Environment.NewLine;
            }
            else
            {
                string unusedArrayAfterFPGATemperature = String.Empty;
                string unusedArrayAfterLongitude = String.Empty;
                string ExposureTimeStr = String.Empty;

                if (fliProCameraManager.DisplayCamera.DeviceType == (int)FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400)
                {
                    ExposureTimeStr = FliProCamera.GetClockFromTime((long)((HDRMode ? (ExposureTime * 2 * 10.26) : ExposureTime * 10.26) / 1000),
                                                       FliProCamera.ExposureUnits_Enum.units_mSecs) + " (" + ExposureTime.ToString() +
                                                       (HDRMode ? " (20.52 µs periods))" : " (10.26 µs periods))");
                }
                else if (fliProCameraManager.DisplayCamera.DeviceType == (int)FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE4040)
                {
                    ExposureTimeStr = FliProCamera.GetClockFromTime((long)(ExposureTime * 10.32 / 1000), FliProCamera.ExposureUnits_Enum.units_mSecs) +
                                                       " (" + ExposureTime.ToString() + " (10.32 µs periods))";
                }
                else if ((fliProCameraManager.DisplayCamera.DeviceType == (int)FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42) ||
                         (fliProCameraManager.DisplayCamera.DeviceType == (int)FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_84))
                {
                    ExposureTimeStr = FliProCamera.GetClockFromTime((long)(ExposureTime / 100), FliProCamera.ExposureUnits_Enum.units_mSecs) +
                                                       " (" + ExposureTime.ToString() + " (10 µs periods))";
                }

                if (UnusedBytesAfterFPGATemperature != null)
                {
                    for (int i = 0; i < UnusedBytesAfterFPGATemperature.Length; i++)
                    {
                        unusedArrayAfterFPGATemperature += (i == 0 ? String.Empty : ", ") + UnusedBytesAfterFPGATemperature[i].ToString("X2");
                    }
                }

                if (UnusedBytesAfterLongitude != null)
                {
                    for (int i = 0; i < UnusedBytesAfterLongitude.Length; i++)
                    {
                        unusedArrayAfterLongitude += (i == 0 ? String.Empty : ", ") + ((i > 0) && (i % 10 == 0) ? Environment.NewLine + "                         " : String.Empty) +
                            UnusedBytesAfterLongitude[i].ToString("X2");
                    }
                }

                if (MagicNumber > 0)
                {
                    value = "MagicNumber           = 0x" + MagicNumber.ToString("X8") + Environment.NewLine +                 // 4	Bytes					x4D657461
                            "MetadataLength        = " + MetadataLength.ToString() + Environment.NewLine +                    // 2    Bytes
                            "MetadataVersion       = " + MetadataVersion.ToString() + Environment.NewLine +                   // 2    Bytes
                            "CameraModel           = " + CameraModel.ToString() + Environment.NewLine +                       // 40   Char
                            "CameraSerialNumber    = " + CameraSerialNumber.ToString() + Environment.NewLine +                // 24   Char
                            "FirmwareVersion       = " + "0x" + FirmwareVersion.ToString("X8") + Environment.NewLine +
                            "TemperatureSetPoint   = " + ((double)TemperatureSetPoint / 10).ToString("0.0") + " (0x" + ((short)TemperatureSetPoint).ToString("X4") + ")" + Environment.NewLine +
                            "CoolerDutyCycle       = " + (CoolerDutyCycle / 2.55).ToString("0.0") + "% (0x" + CoolerDutyCycle.ToString("X2") + ")" + Environment.NewLine +                  // 1	Bytes

                            //"CaptureYear           = " + CaptureYear.ToString() + Environment.NewLine +                       // 1	Bytes                   Value = (7:1) Year-2000 , (0) Month MSb			
                            //"CaptureMonth          = " + CaptureMonth.ToString() + Environment.NewLine +                      // 1	Bytes                   Value = (7:5) Month, (4:0) Day	
                            //"CaptureDay            = " + CaptureDay.ToString() + Environment.NewLine +
                            //"CaptureTime           = " + FliProCamera.GetClockFromTime(CaptureTime / 10000, FliProCamera.ExposureUnits_Enum.units_mSecs) + " ("
                            //                           + CaptureTime.ToString() + " (100ns periods)" + Environment.NewLine +       // 5	Bytes                   Value = (39:0) Time in 100ns periods				ISO-8601
                            "SensorPixelBitDepth   = " + SensorPixelBitDepth.ToString() + Environment.NewLine +                                                    // 1	Bytes                   Value = Bit Depth	
                            "DataPixelBitDepth     = " + DataPixelBitDepth.ToString() + " (value stored as value-1)" + Environment.NewLine +                       // 1	Bytes                   Value = (Bit Depth - 1) (3:0),  ZeroPoint(7:4)	
                            "DataZeroPoint         = " + DataZeroPoint.ToString() + Environment.NewLine +
                            "HorizontalPixels      = " + HorizontalPixels.ToString() + " (0x" + HorizontalPixels.ToString("X4") + ")" + Environment.NewLine +      // 2	Bytes
                            "VerticalPixels        = " + VerticalPixels.ToString() + " (0x" + VerticalPixels.ToString("X4") + ")" + Environment.NewLine +          // 2	Bytes
                            "ExposureTime          = " + ExposureTimeStr + Environment.NewLine +                                                                   // 4	Bytes                   Value = Exposure Time in 10us periods	
                            "SensorColdTemp        = " + ((double)SensorColdTemp / 10).ToString("0.0") + " (0x" + ((short)SensorColdTemp).ToString("X4") + ")" + Environment.NewLine +          // 2	Bytes
                            "SensorTemp            = " + GetConvertedTemperature(SensorTemp) + " (0x" + ((short)SensorTemp).ToString("X4") + ")" + Environment.NewLine +                  // 2	Bytes
                            "GlobalGain            = " + GlobalGain.ToString() + " (0x" + GlobalGain.ToString("X2") + ")" + Environment.NewLine +                  // 1	Bytes
                            "LowGain               = " + LowGain.ToString() + " (0x" + LowGain.ToString("X2") + ")" + Environment.NewLine +                        // 1	Bytes                   Value = (2:0) for 400 Sensor	
                            "HighGain               = " + HighGain.ToString() + " (0x" + HighGain.ToString("X2") + ")" + Environment.NewLine +                        // 1	Bytes                   Value = (2:0) for 400 Sensor		
                            "ReferencePixelsPerRow = " + DummyPixelsPerRow.ToString() + " (0x" + DummyPixelsPerRow.ToString("X2") + ")" + Environment.NewLine +    // 1	Bytes
                            "UnusedByte            = " + UnusedByteAfterDummyPixelsPerRow.ToString() + " (0x" + UnusedByteAfterDummyPixelsPerRow.ToString("X2") + ")" + Environment.NewLine +                  // 1	Bytes
                            "ControlBlock          = " + ControlBlock.ToString() + " (0x" + ControlBlock.ToString("X2") + ")" + Environment.NewLine + Environment.NewLine +            // 1	Bytes

                            "DeadPixelCorrection   = " + DeadPixelCorrection.ToString() + Environment.NewLine +               // 7								
                            "PixelOrderedImage     = " + PixelOrderedImage.ToString() + Environment.NewLine +                 // 6								
                            "HDRMode               = " + HDRMode.ToString() + Environment.NewLine +                           // 5								
                            "LowNoise              = " + LowNoise.ToString() + Environment.NewLine +                          // 4								
                            "LowDarkCurrent        = " + LowDarkCurrent.ToString() + Environment.NewLine +                    // 3								
                            "HorzScanDirInvert     = " + HorizontalScanDirectionInvert.ToString() + Environment.NewLine +     // 2								
                            "VertScanDirInvert     = " + VerticalScanDirectionInvert.ToString() + Environment.NewLine +       // 1	
                            "NonRowAllignedImage   = " + NonRowAllignedImage.ToString() + Environment.NewLine + Environment.NewLine +             // 0	

                            "HorizontalBinning     = " + HorizontalBinning.ToString() + " (0x" + HorizontalBinning.ToString("X2") + ")" + Environment.NewLine +                        // 1	Bytes									
                            "VerticalBinning       = " + VerticalBinning.ToString() + " (0x" + VerticalBinning.ToString("X2") + ")" + Environment.NewLine +                            // 1	Bytes									
                            "BlackLevelAdjust      = " + BlackLevelAdjust.ToString() + " (0x" + BlackLevelAdjust.ToString("X4") + ")" + Environment.NewLine +                          // 3	Bytes				    Value = (13:0) for 400 Sensor				
                            "BlackLevelSun         = " + BlackLevelSun.ToString() + " (0x" + BlackLevelSun.ToString("X4") + ")" + Environment.NewLine +                                // 2	Bytes					Value = (5:0) for 400 Sensor				
                            "ShutterOpenDelay      = " + ShutterOpenDelay.ToString() + " (0x" + ((short)ShutterOpenDelay).ToString("X4") + ")" + Environment.NewLine +                 // 2	Bytes					Value = Signed				
                            "ShutterCloseDelay     = " + ShutterCloseDelay.ToString() + " (0x" + ((short)ShutterCloseDelay).ToString("X4") + ")" + Environment.NewLine +               // 2	Bytes					Value = Signed				
                            "IlluminStartDelay     = " + IlluminationStartDelay.ToString() + " (0x" + ((short)IlluminationStartDelay).ToString("X4") + ")" + Environment.NewLine +     // 2	Bytes					Value = Signed				
                            "IlluminStopDelay      = " + IlluminationStopDelay.ToString() + " (0x" + ((short)IlluminationStopDelay).ToString("X4") + ")" + Environment.NewLine +       // 2	Bytes					Value = Signed				
                            "TrackingStartRow      = " + TrackingStartRow.ToString() + " (0x" + TrackingStartRow.ToString("X4") + ")" + Environment.NewLine +                          // 4	Bytes					Value = (31:16) Start Row, (15:0) Start Column
                            "TrackingStartColumn   = " + TrackingStartColumn.ToString() + " (0x" + TrackingStartColumn.ToString("X4") + ")" + Environment.NewLine +
                            "TrackingStopRow       = " + TrackingStopRow.ToString() + " (0x" + TrackingStopRow.ToString("X4") + ")" + Environment.NewLine +                            // 4	Bytes					Value = (31:16) Stop Row, (15:0) Stop Column
                            "TrackingStopColumn    = " + TrackingStopColumn.ToString() + " (0x" + TrackingStopColumn.ToString("X4") + ")" + Environment.NewLine +
                            "ImageStartRow         = " + ImageStartRow.ToString() + " (0x" + ImageStartRow.ToString("X4") + ")" + Environment.NewLine +                                // 4	Bytes					Value = (31:16) Start Row, (15:0) Start Column	
                            "ImageStartColumn      = " + ImageStartColumn.ToString() + " (0x" + ImageStartColumn.ToString("X4") + ")" + Environment.NewLine +
                            "ImageStopRow          = " + ImageStopRow.ToString() + " (0x" + ImageStopRow.ToString("X4") + ")" + Environment.NewLine +                                  // 4	Bytes					Value = (31:16) Stop Row, (15:0) Stop Column
                            "ImageStopColumn       = " + ImageStopColumn.ToString() + " (0x" + ImageStopColumn.ToString("X4") + ")" + Environment.NewLine +
                            "TrackingFramesPerImageFrame = " + TrackingFramesPerImageFrame.ToString() + " (0x" + TrackingFramesPerImageFrame.ToString("X4") + ")" + Environment.NewLine +     // 2	Bytes									
                            "FrameNumber           = " + FrameNumber.ToString() + " (0x" + FrameNumber.ToString("X4") + ")" + Environment.NewLine +                                    // 2	Bytes
                            "StartingExposureRow   = " + StartingExposureRow.ToString() + " (0x" + StartingExposureRow.ToString("X4") + ")" + Environment.NewLine +                    // 2	Bytes
                            "PreReferenceRows      = " + PreDummyRows.ToString() + Environment.NewLine +                                                                          // 3	Bytes                   Value = (23:12) Pre-Rows (12 bits), (11:0) Post-Rows (12 bits)		
                            "PostReferenceRows     = " + PostDummyRows.ToString() + Environment.NewLine +
                            "NumberOfDataChannels  = " + NumberOfDataChannels.ToString() + Environment.NewLine +                                                                 // 1	Bytes
                            "ControlBlock2         = " + ControlBlock2.ToString() + " (0x" + ControlBlock2.ToString("X2") + ")" + Environment.NewLine + Environment.NewLine +                              // 1	Bytes

                            "BackSideIlluminated   = " + BackSideIlluminated.ToString() + Environment.NewLine +                           // 7
                            "GPSLocked             = " + GPSLocked.ToString() + Environment.NewLine + Environment.NewLine +               // 6

                            "BaseTemperature       = " + BaseTemperature.ToString() + Environment.NewLine +                                                                 // 1	Bytes
                            "FPGATemperature       = " + FPGATemperature.ToString() + Environment.NewLine +                                                                 // 1	Bytes

                            "UnusedBytes           = [" + unusedArrayAfterFPGATemperature + "]" + Environment.NewLine + Environment.NewLine +                           // 2

                            "ControlBlockTest      = " + ControlBlockTest.ToString() + " (0x" + ControlBlockTest.ToString("X2") + ")" + Environment.NewLine + Environment.NewLine +  // 1	Bytes

                            "TestUseShiftedAveraging = " + UseShiftedAveraging.ToString() + Environment.NewLine +                                            // 7
                            "TestIsStackedFrame    = " + IsStackedFrame.ToString() + Environment.NewLine +                                                   // 6
                            "TestIsHighFrame       = " + IsHighFrame.ToString() + Environment.NewLine +                                                      // 5
                            "TestIsMergedFrame     = " + IsMergedFrame.ToString() + Environment.NewLine + Environment.NewLine +                              // 4

                            "TestLowGainAbsolute   = " + GetGainAbsoluteString(TestLowGainAbsolute) + Environment.NewLine +                                  // 2	Bytes
                            "TestHighGainAbsolute  = " + GetGainAbsoluteString(TestHighGainAbsolute) + Environment.NewLine + Environment.NewLine +           // 2	Bytes

                            "CaptureDateString     = " + CaptureString + Environment.NewLine +
                            "CaptureDate           = " + CaptureMonth.ToString("00") + "/" + CaptureDay.ToString("00") + "/" + CaptureYear.ToString("0000") +
                                                         " " + CaptureHour.ToString("00") + ":" + CaptureMinute.ToString("00") + ":" + CaptureSecond.ToString("00") +
                                                         "." + CaptureNanoseconds.ToString("000000000") + Environment.NewLine +
                            "Latitude              = " + LatitudeDegrees.ToString() + "°" + LatitudeMinutes.ToString("0.000") + "' " + (LatitudeNorth ? "N" : "S") + Environment.NewLine +
                            "Longitude             = " + LongitudeDegrees.ToString() + "°" + LongitudeMinutes.ToString("0.000") + "' " + (LongitudeEast ? "E" : "W") + Environment.NewLine +
                            "FliPilotVersion       = " + FliPilotVersion + Environment.NewLine +
                            "LibFliProVersion      = " + LibFliProVersion + Environment.NewLine +
                            "HorizontalPixelSize   = " + ((double)HorizontalPixelSize / 10).ToString("0.0") + " (0x" + HorizontalPixelSize.ToString("X2") + ") - (measured in 0.1 µm)" + Environment.NewLine +
                            "VerticalPixelSize     = " + ((double)VerticalPixelSize / 10).ToString("0.0") + " (0x" + VerticalPixelSize.ToString("X2") + ") - (measured in 0.1 µm)" + Environment.NewLine +
                            "UnusedBytes           = " + "[" + unusedArrayAfterLongitude + "]" + Environment.NewLine + Environment.NewLine;
                }
                else
                {
                    value = "Metadata is Invalid.";
                }
            }

            return value;
        }

        public double CreateDirection(uint rawData)
        {
            double location = (rawData & DIRECTION_DEGREES_MASK);

            location /= DIRECTION_DEGREES_DIVISOR;

            if ((rawData & DIRECTION_FLAG_MASK) == 0)
            {
                location *= -1;
            }

            return location;
        }

        public byte[] GetByteArray()
        {
            int currentIndex = 0;
            int signedValue = 0;
            byte[] metadata = new byte[MetadataLength];
            byte[] dataArray;

            StoreValue(MAGIC_NUMBER, metadata, ref currentIndex, MAGIC_NUMBER_SIZE);
            StoreValue(MetadataLength, metadata, ref currentIndex, METADATA_LENGTH_SIZE);
            StoreValue(MetadataVersion, metadata, ref currentIndex, METADATA_VERSION_SIZE);
            if (String.IsNullOrEmpty(CameraModel))
            {
                CameraModel = "No Camera Model Found";
            }
            Array.Copy(Encoding.ASCII.GetBytes(CameraModel), 0, metadata, currentIndex, CameraModel.Length);
            currentIndex += CAMERA_MODEL_SIZE;
            if (String.IsNullOrEmpty(CameraSerialNumber))
            {
                CameraSerialNumber = "Unknown";
            }
            Array.Copy(Encoding.ASCII.GetBytes(CameraSerialNumber), 0, metadata, currentIndex, CameraSerialNumber.Length);
            currentIndex += CAMERA_SERIAL_NUMBER_SIZE;

            if (MetadataVersion == META_DATA_VERSION_1)
            {
                ushort dateValue = (ushort)(CaptureYear - 2000);
                dateValue = (ushort)((dateValue << 9) + (CaptureMonth << 5) + CaptureDay);

                // BitConverter.GetBytes swaps the bytes so store them reversed
                dataArray = BitConverter.GetBytes(dateValue);
                metadata[currentIndex] = dataArray[1];
                currentIndex++;
                metadata[currentIndex] = dataArray[0];
                currentIndex++;

                StoreValue(CaptureTime, metadata, ref currentIndex, CAPTURE_TIME_SIZE + 3);
                currentIndex -= 3;   // Need to back up the 3 extra we just added

                metadata[currentIndex] = (byte)SensorPixelBitDepth;
                currentIndex += SENSOR_PIXEL_BIT_DEPTH_SIZE;

                metadata[currentIndex] = (byte)((DataZeroPoint << 4) + (DataPixelBitDepth - 1));
                currentIndex += DATA_PIXEL_BIT_DEPTH_ZERO_POINT_SIZE;

                StoreValue(HorizontalPixels, metadata, ref currentIndex, HORIZONTAL_PIXELS_SIZE);
                StoreValue(VerticalPixels, metadata, ref currentIndex, VERTICAL_PIXELS_SIZE);
                StoreValue(ExposureTime, metadata, ref currentIndex, EXPOSURE_TIME_SIZE);
                StoreValue((SensorColdTemp & 0xFFFF), metadata, ref currentIndex, SENSOR_BASE_TEMP_SIZE);
                StoreValue((SensorTemp & 0xFFFF), metadata, ref currentIndex, SENSOR_TEMP_SIZE);

                metadata[currentIndex] = (byte)GlobalGain;
                currentIndex += LDR_GAIN_SIZE;

                metadata[currentIndex] = (byte)LowGain;
                currentIndex += LDR_GAIN_SIZE;

                metadata[currentIndex] = (byte)HighGain;
                currentIndex += HDR_GAIN_SIZE;

                metadata[currentIndex] = (byte)DummyPixelsPerRow;
                currentIndex += DUMMY_PIXELS_PER_ROW_SIZE;

                metadata[currentIndex] = (byte)UnusedByteAfterDummyPixelsPerRow;
                currentIndex += UNUSED_BYTE_AFTER_DUMMY_PIXELS_PER_ROW_SIZE;

                metadata[currentIndex] = (byte)((DeadPixelCorrection ? DEAD_PIXEL_CORRECTION_MASK : 0) +
                                                (PixelOrderedImage ? PIXEL_ORDERED_IMAGE_MASK : 0) +
                                                (HDRMode ? HDR_MODE : 0) +
                                                (LowNoise ? LOW_NOISE_MASK : 0) +
                                                (LowDarkCurrent ? LOW_DARK_CURRENT_MASK : 0) +
                                                (HorizontalScanDirectionInvert ? HORIZONTAL_SCAN_DIRECTION_INVERT_MASK : 0) +
                                                (VerticalScanDirectionInvert ? VERTICAL_SCAN_DIRECTION_INVERT_MASK : 0) +
                                                (NonRowAllignedImage ? NON_ROW_ALLIGNED_IMAGE_MASK : 0));
                currentIndex += CONTROL_BLOCK_SIZE;

                metadata[currentIndex] = (byte)HorizontalBinning;
                currentIndex += HORIZONTAL_BINNING_SIZE;

                metadata[currentIndex] = (byte)VerticalBinning;
                currentIndex += VERTICAL_BINNING_SIZE;

                StoreValue(BlackLevelAdjust, metadata, ref currentIndex, BLACK_LEVEL_ADJUST_SIZE + 1);  // Add 1 to get the function to work properly
                currentIndex--;                                                                         // Then subtract it again to put us back on the right count
                StoreValue(BlackLevelSun, metadata, ref currentIndex, BLACK_LEVEL_SUN_SIZE);

                signedValue = ShutterOpenDelay;
                if (signedValue < 0)
                {
                    // Convert to Sign Magnitude
                    signedValue *= -1;
                    signedValue += NEGATIVE_BYTE_MASK;
                }
                StoreValue(signedValue, metadata, ref currentIndex, SHUTTER_OPEN_DELAY_SIZE);

                signedValue = ShutterCloseDelay;
                if (signedValue < 0)
                {
                    // Convert to Sign Magnitude
                    signedValue *= -1;
                    signedValue += NEGATIVE_BYTE_MASK;
                }
                StoreValue(signedValue, metadata, ref currentIndex, SHUTTER_CLOSE_DELAY_SIZE);

                signedValue = IlluminationStartDelay;
                if (signedValue < 0)
                {
                    // Convert to Sign Magnitude
                    signedValue *= -1;
                    signedValue += NEGATIVE_BYTE_MASK;
                }
                StoreValue(signedValue, metadata, ref currentIndex, ILLUMINATION_START_DELAY_SIZE);

                signedValue = IlluminationStopDelay;
                if (signedValue < 0)
                {
                    // Convert to Sign Magnitude
                    signedValue *= -1;
                    signedValue += NEGATIVE_BYTE_MASK;
                }
                StoreValue(signedValue, metadata, ref currentIndex, ILLUMINATION_STOP_DELAY_SIZE);

                StoreValue(TrackingStartRow, metadata, ref currentIndex, TRACKING_START_ROW_SIZE);
                StoreValue(TrackingStartColumn, metadata, ref currentIndex, TRACKING_START_COLUMN_SIZE);
                StoreValue(TrackingStopRow, metadata, ref currentIndex, TRACKING_STOP_ROW_SIZE);
                StoreValue(TrackingStopColumn, metadata, ref currentIndex, TRACKING_STOP_COLUMN_SIZE);
                StoreValue(ImageStartRow, metadata, ref currentIndex, IMAGE_START_ROW_SIZE);
                StoreValue(ImageStartColumn, metadata, ref currentIndex, IMAGE_START_COLUMN_SIZE);
                StoreValue(ImageStopRow, metadata, ref currentIndex, IMAGE_STOP_ROW_SIZE);
                StoreValue(ImageStopColumn, metadata, ref currentIndex, IMAGE_STOP_COLUMN_SIZE);
                StoreValue(TrackingFramesPerImageFrame, metadata, ref currentIndex, TRACKING_FRAMES_PER_IMAGE_FRAME_SIZE);
                StoreValue(FrameNumber, metadata, ref currentIndex, FRAME_NUMBER_SIZE);
                StoreValue(StartingExposureRow, metadata, ref currentIndex, STARTING_EXPOSURE_ROW_SIZE);

                metadata[currentIndex] = (byte)(PreDummyRows & 0xFF);
                metadata[currentIndex + 1] = (byte)(((PreDummyRows & 0x0F00) >> 8) | ((PostDummyRows & 0x0F) << 4));
                metadata[currentIndex + 2] = (byte)((PostDummyRows & 0x0FF0) >> 4);
                currentIndex += PRE_DUMMY_ROWS_SIZE;

                metadata[currentIndex] = (byte)NumberOfDataChannels;
                currentIndex += NUMBER_OF_CHANNELS_SIZE;

                metadata[currentIndex] = 0xFF;
                metadata[currentIndex] = (byte)(metadata[currentIndex] & (BackSideIlluminated ? BACK_SIDE_ILLUMINATED_ON_MASK : BACK_SIDE_ILLUMINATED_OFF_MASK));
                currentIndex += CONTROL_BLOCK_SIZE;

                StoreValue(BaseTemperature, metadata, ref currentIndex, BASE_TEMPERATURE_SIZE);
                StoreValue(FPGATemperature, metadata, ref currentIndex, FPGA_TEMPERATURE_SIZE);

                Array.Copy(UnusedBytesAfterFPGATemperature, 0, metadata, currentIndex, UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE);
                currentIndex += UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE;

                metadata[currentIndex] = 0xFF;
                metadata[currentIndex] -= (byte)(UseShiftedAveraging ? TEST_USE_SHIFTED_AVERAGING_MASK : 0);
                metadata[currentIndex] -= (byte)(IsStackedFrame ? TEST_IS_STACKED_FRAME_MASK : 0);
                metadata[currentIndex] -= (byte)(IsHighFrame ? TEST_IS_HIGH_FRAME_MASK : 0);
                metadata[currentIndex] -= (byte)(IsMergedFrame ? TEST_IS_MERGED_FRAME_MASK : 0);
                currentIndex += CONTROL_BLOCK_SIZE;

                StoreValue(TestLowGainAbsolute, metadata, ref currentIndex, TEST_LOW_GAIN_ABSOLUTE_SIZE);
                StoreValue(TestHighGainAbsolute, metadata, ref currentIndex, TEST_HIGH_GAIN_ABSOLUTE_SIZE);

                StoreValue(FirmwareVersion, metadata, ref currentIndex, FIRMWARE_VERSION_SIZE);
            }
            else
            {
                StoreValue(FirmwareVersion, metadata, ref currentIndex, FIRMWARE_VERSION_SIZE);

                StoreValue((TemperatureSetPoint & 0xFFFF), metadata, ref currentIndex, TEMPERATURE_SET_POINT_SIZE);

                metadata[currentIndex] = CoolerDutyCycle;
                currentIndex += COOLER_DUTY_CYCLE_SIZE;

                metadata[currentIndex] = (byte)SensorPixelBitDepth;
                currentIndex += SENSOR_PIXEL_BIT_DEPTH_SIZE;

                metadata[currentIndex] = (byte)((DataZeroPoint << 4) + (DataPixelBitDepth - 1));
                currentIndex += DATA_PIXEL_BIT_DEPTH_ZERO_POINT_SIZE;

                StoreValue(HorizontalPixels, metadata, ref currentIndex, HORIZONTAL_PIXELS_SIZE);
                StoreValue(VerticalPixels, metadata, ref currentIndex, VERTICAL_PIXELS_SIZE);
                StoreValue(ExposureTime, metadata, ref currentIndex, EXPOSURE_TIME_SIZE);
                StoreValue((SensorColdTemp & 0xFFFF), metadata, ref currentIndex, SENSOR_BASE_TEMP_SIZE);
                StoreValue((SensorTemp & 0xFFFF), metadata, ref currentIndex, SENSOR_TEMP_SIZE);

                metadata[currentIndex] = (byte)GlobalGain;
                currentIndex += LDR_GAIN_SIZE;

                metadata[currentIndex] = (byte)LowGain;
                currentIndex += LDR_GAIN_SIZE;

                metadata[currentIndex] = (byte)HighGain;
                currentIndex += HDR_GAIN_SIZE;

                metadata[currentIndex] = (byte)DummyPixelsPerRow;
                currentIndex += DUMMY_PIXELS_PER_ROW_SIZE;

                metadata[currentIndex] = (byte)UnusedByteAfterDummyPixelsPerRow;
                currentIndex += UNUSED_BYTE_AFTER_DUMMY_PIXELS_PER_ROW_SIZE;

                metadata[currentIndex] = (byte)((DeadPixelCorrection ? DEAD_PIXEL_CORRECTION_MASK : 0) +
                                                (PixelOrderedImage ? PIXEL_ORDERED_IMAGE_MASK : 0) +
                                                (HDRMode ? HDR_MODE : 0) +
                                                (LowNoise ? LOW_NOISE_MASK : 0) +
                                                (LowDarkCurrent ? LOW_DARK_CURRENT_MASK : 0) +
                                                (HorizontalScanDirectionInvert ? HORIZONTAL_SCAN_DIRECTION_INVERT_MASK : 0) +
                                                (VerticalScanDirectionInvert ? VERTICAL_SCAN_DIRECTION_INVERT_MASK : 0) +
                                                (NonRowAllignedImage ? NON_ROW_ALLIGNED_IMAGE_MASK : 0));
                currentIndex += CONTROL_BLOCK_SIZE;

                metadata[currentIndex] = (byte)HorizontalBinning;
                currentIndex += HORIZONTAL_BINNING_SIZE;

                metadata[currentIndex] = (byte)VerticalBinning;
                currentIndex += VERTICAL_BINNING_SIZE;

                StoreValue(BlackLevelAdjust, metadata, ref currentIndex, BLACK_LEVEL_ADJUST_SIZE + 1);  // Add 1 to get the function to work properly
                currentIndex--;                                                                         // Then subtract it again to put us back on the right count
                StoreValue(BlackLevelSun, metadata, ref currentIndex, BLACK_LEVEL_SUN_SIZE);

                signedValue = ShutterOpenDelay;
                if (signedValue < 0)
                {
                    // Convert to Sign Magnitude
                    signedValue *= -1;
                    signedValue += NEGATIVE_BYTE_MASK;
                }
                StoreValue(signedValue, metadata, ref currentIndex, SHUTTER_OPEN_DELAY_SIZE);

                signedValue = ShutterCloseDelay;
                if (signedValue < 0)
                {
                    // Convert to Sign Magnitude
                    signedValue *= -1;
                    signedValue += NEGATIVE_BYTE_MASK;
                }
                StoreValue(signedValue, metadata, ref currentIndex, SHUTTER_CLOSE_DELAY_SIZE);

                signedValue = IlluminationStartDelay;
                if (signedValue < 0)
                {
                    // Convert to Sign Magnitude
                    signedValue *= -1;
                    signedValue += NEGATIVE_BYTE_MASK;
                }
                StoreValue(signedValue, metadata, ref currentIndex, ILLUMINATION_START_DELAY_SIZE);

                signedValue = IlluminationStopDelay;
                if (signedValue < 0)
                {
                    // Convert to Sign Magnitude
                    signedValue *= -1;
                    signedValue += NEGATIVE_BYTE_MASK;
                }
                StoreValue(signedValue, metadata, ref currentIndex, ILLUMINATION_STOP_DELAY_SIZE);

                StoreValue(TrackingStartRow, metadata, ref currentIndex, TRACKING_START_ROW_SIZE);
                StoreValue(TrackingStartColumn, metadata, ref currentIndex, TRACKING_START_COLUMN_SIZE);
                StoreValue(TrackingStopRow, metadata, ref currentIndex, TRACKING_STOP_ROW_SIZE);
                StoreValue(TrackingStopColumn, metadata, ref currentIndex, TRACKING_STOP_COLUMN_SIZE);
                StoreValue(ImageStartRow, metadata, ref currentIndex, IMAGE_START_ROW_SIZE);
                StoreValue(ImageStartColumn, metadata, ref currentIndex, IMAGE_START_COLUMN_SIZE);
                StoreValue(ImageStopRow, metadata, ref currentIndex, IMAGE_STOP_ROW_SIZE);
                StoreValue(ImageStopColumn, metadata, ref currentIndex, IMAGE_STOP_COLUMN_SIZE);
                StoreValue(TrackingFramesPerImageFrame, metadata, ref currentIndex, TRACKING_FRAMES_PER_IMAGE_FRAME_SIZE);
                StoreValue(FrameNumber, metadata, ref currentIndex, FRAME_NUMBER_SIZE);
                StoreValue(StartingExposureRow, metadata, ref currentIndex, STARTING_EXPOSURE_ROW_SIZE);

                metadata[currentIndex] = (byte)(PreDummyRows & 0xFF);
                metadata[currentIndex + 1] = (byte)(((PreDummyRows & 0x0F00) >> 8) | ((PostDummyRows & 0x0F) << 4));
                metadata[currentIndex + 2] = (byte)((PostDummyRows & 0x0FF0) >> 4);
                currentIndex += PRE_DUMMY_ROWS_SIZE;

                metadata[currentIndex] = (byte)NumberOfDataChannels;
                currentIndex += NUMBER_OF_CHANNELS_SIZE;

                metadata[currentIndex] = 0xFF;
                metadata[currentIndex] = (byte)(metadata[currentIndex] & (BackSideIlluminated ? BACK_SIDE_ILLUMINATED_ON_MASK : BACK_SIDE_ILLUMINATED_OFF_MASK));
                currentIndex += CONTROL_BLOCK_SIZE;

                StoreValue(BaseTemperature, metadata, ref currentIndex, BASE_TEMPERATURE_SIZE);
                StoreValue(FPGATemperature, metadata, ref currentIndex, FPGA_TEMPERATURE_SIZE);

                Array.Copy(UnusedBytesAfterFPGATemperature, 0, metadata, currentIndex, UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE);
                currentIndex += UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE;

                metadata[currentIndex] = 0xFF;
                metadata[currentIndex] -= (byte)(UseShiftedAveraging ? TEST_USE_SHIFTED_AVERAGING_MASK : 0);
                metadata[currentIndex] -= (byte)(IsStackedFrame ? TEST_IS_STACKED_FRAME_MASK : 0);
                metadata[currentIndex] -= (byte)(IsHighFrame ? TEST_IS_HIGH_FRAME_MASK : 0);
                metadata[currentIndex] -= (byte)(IsMergedFrame ? TEST_IS_MERGED_FRAME_MASK : 0);
                currentIndex += CONTROL_BLOCK_SIZE;

                StoreValue(TestLowGainAbsolute, metadata, ref currentIndex, TEST_LOW_GAIN_ABSOLUTE_SIZE);
                StoreValue(TestHighGainAbsolute, metadata, ref currentIndex, TEST_HIGH_GAIN_ABSOLUTE_SIZE);

                Array.Copy(Encoding.ASCII.GetBytes(CaptureString), 0, metadata, currentIndex, CaptureString.Length);
                currentIndex += TIMESTAMP_SIZE;

                StoreValue(LatitudeRaw, metadata, ref currentIndex, LATITUDE_SIZE);
                StoreValue(LongitudeRaw, metadata, ref currentIndex, LONGITUDE_SIZE);

                Array.Copy(UnusedBytesAfterLongitude, 0, metadata, currentIndex, UnusedBytesAfterLongitude.Length);
                currentIndex += UNUSED_BYTES_AFTER_LONGITUDE_SIZE;
            }

            return metadata;
        }

        public static string GetCameraModelSpaces(string name)
        {
            string spaces = String.Empty;

            for (int i = 0; i < CAMERA_MODEL_SIZE - name.Length; i++)
            {
                spaces += " ";
            }
            return spaces;
        }

        public static string GetCameraSerialNumber(string name)
        {
            string spaces = String.Empty;

            for (int i = 0; i < CAMERA_SERIAL_NUMBER_SIZE - name.Length; i++)
            {
                spaces += " ";
            }
            return spaces;
        }

        /// <summary>
        /// Converts the sensor temperature to a string
        /// </summary>
        /// <param name="sensorTemp"></param>
        /// <returns></returns>
        public string GetConvertedTemperature(int sensorTemp)
        {
            int signed = ((sensorTemp & NEGATIVE_SHORT_MASK) == NEGATIVE_SHORT_MASK ? -1 : 1);
            double temperature = ((double)(sensorTemp & NEGATIVE_SHORT_CLEAR_MASK) * signed) / 10;
            return temperature.ToString("0.0");
        }

        /// <summary>
        /// Converts the converted sensor temperature string back to the original sensor temperature
        /// </summary>
        /// <param name="sensorTemp"></param>
        /// <returns></returns>
        public int GetConvertedTemperature(string sensorTempStr)
        {
            double temperature = 0;

            try
            {
                temperature = Convert.ToDouble(sensorTempStr);
                temperature = (temperature * 2.0209) + 617.45;
            }
            catch { }

            return (int)Math.Round(temperature, 0);
        }

        public string GetCaptureTime()
        {
            if ((MetadataVersion == META_DATA_VERSION_1) || (String.IsNullOrEmpty(CaptureString)))
            {
                double newValue = CaptureTime / 10000;  // convert the time to milliseconds
                long hours = (long)(newValue / MILLISECS_PER_HOUR);
                newValue -= hours * MILLISECS_PER_HOUR;
                long mins = (long)(newValue / MILLISECS_PER_MIN);
                newValue -= mins * MILLISECS_PER_MIN;
                long secs = (long)(newValue / MILLISECS_PER_SEC);
                newValue -= secs * MILLISECS_PER_SEC;
                DateTime date = new DateTime(CaptureYear, CaptureMonth, CaptureDay, (int)hours, (int)mins, (int)secs, (int)newValue);
                CaptureString = date.ToString("yyyy-MM-ddTHH:mm:ss.fff000");
            }
            return CaptureString;
        }

        public string GetGainAbsoluteString(int gainAbsolute)
        {
            string gainStr = ((double)gainAbsolute / 100).ToString("0.000");
            return gainStr;
        }

        public static int GetMetadataLengthFromArray(byte[] rawData)
        {
            int currentIndex = 0;
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];
            int metadataLength = 0;
            int magicNumber = 0;

            if (rawData.Length > 6)
            {
                Regex regex = new Regex(@"[^a-zA-Z0-9\s\-]", (RegexOptions)0);

                Array.Copy(rawData, currentIndex, fourBytes, 0, MAGIC_NUMBER_SIZE);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(fourBytes);
                }
                magicNumber = BitConverter.ToInt32(fourBytes, 0);
                currentIndex += MAGIC_NUMBER_SIZE;

                Array.Copy(rawData, currentIndex, twoBytes, 0, 2);
                if (BitConverter.IsLittleEndian)
                {
                    Array.Reverse(twoBytes);
                }
                metadataLength = BitConverter.ToUInt16(twoBytes, 0);
            }
            return metadataLength;
        }

        public void SetPaddedModel(string model)
        {
            CameraModel = model;
            for (int i = model.Length; i < CAMERA_MODEL_SIZE; i++)
            {
                CameraModel += " ";
            }
        }

        public void SetPaddedSerialNumber(string serialNumber)
        {
            CameraSerialNumber = serialNumber;
            for (int i = serialNumber.Length; i < CAMERA_SERIAL_NUMBER_SIZE; i++)
            {
                CameraSerialNumber += " ";
            }
        }

        #endregion Public Functions

        //**************************************************************
        #region Private Functions

        private void FixDateAndTime(byte[] metadata, int currentIndex, DateTime captureDatetime)
        {
            ushort dateValue = (ushort)(captureDatetime.Year - 2000);
            dateValue = (ushort)((dateValue << 9) + (captureDatetime.Month << 5) + captureDatetime.Day);

            byte[] dataArray = BitConverter.GetBytes(dateValue);
            metadata[currentIndex] = dataArray[1];
            currentIndex++;
            metadata[currentIndex] = dataArray[0];
            currentIndex++;

            ulong targetLong = ((ulong)((captureDatetime.Hour * MILLISECS_PER_HOUR) + (captureDatetime.Minute * MILLISECS_PER_MIN) +
                        (captureDatetime.Second * MILLISECS_PER_SEC) + captureDatetime.Millisecond)) * 10000;
            dataArray = BitConverter.GetBytes(targetLong);
            Array.Reverse(dataArray);
            Array.Copy(dataArray, 3, metadata, currentIndex, CAPTURE_TIME_SIZE);
        }

        private void ParseMetadataVersion1(byte[] metadata, int currentIndex, DateTime captureDatetime)
        {
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];
            byte[] eightBytes = new byte[8] { 0, 0, 0, 0, 0, 0, 0, 0 };
            int signed = 1;
            Regex regex = new Regex(@"[^a-zA-Z0-9\s\-]", (RegexOptions)0);

            CameraModel = System.Text.Encoding.UTF8.GetString(metadata, currentIndex, CAMERA_MODEL_SIZE);
            CameraModel = regex.Replace(CameraModel, "").Trim();
            currentIndex += CAMERA_MODEL_SIZE;
            CameraSerialNumber = System.Text.Encoding.UTF8.GetString(metadata, currentIndex, CAMERA_SERIAL_NUMBER_SIZE);
            CameraSerialNumber = regex.Replace(CameraSerialNumber, "").Trim();
            currentIndex += CAMERA_SERIAL_NUMBER_SIZE;

            CaptureYear = (metadata[currentIndex] >> 1) + 2000;
            if (CaptureYear < 2018)
            {
                FixDateAndTime(metadata, currentIndex, captureDatetime);
                CaptureYear = (metadata[currentIndex] >> 1) + 2000;
            }
            CaptureMonth = ((metadata[currentIndex] & 1) << 3) + (metadata[currentIndex + 1] >> 5);
            CaptureDay = (metadata[currentIndex + 1] & 0x1F);
            currentIndex += (CAPTURE_YEAR_SIZE + CAPTURE_MONTH_SIZE);

            eightBytes[0] = 0;
            eightBytes[1] = 0;
            eightBytes[2] = 0;
            Array.Copy(metadata, currentIndex, eightBytes, 3, CAPTURE_TIME_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(eightBytes);
            }
            CaptureTime = BitConverter.ToInt64(eightBytes, 0);
            currentIndex += CAPTURE_TIME_SIZE;
            
            SensorPixelBitDepth = metadata[currentIndex];
            currentIndex += SENSOR_PIXEL_BIT_DEPTH_SIZE;

            DataPixelBitDepth = (metadata[currentIndex] & 0x0F) + 1;   // NOTE: DataPixelBitDepth is stored as -1
            DataZeroPoint = (metadata[currentIndex] & 0xF0) >> 4;
            currentIndex += DATA_PIXEL_BIT_DEPTH_ZERO_POINT_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            HorizontalPixels = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += HORIZONTAL_PIXELS_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            VerticalPixels = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += VERTICAL_PIXELS_SIZE;

            Array.Copy(metadata, currentIndex, fourBytes, 0, 4);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            ExposureTime = BitConverter.ToInt32(fourBytes, 0);
            currentIndex += EXPOSURE_TIME_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            SensorColdTemp = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += SENSOR_BASE_TEMP_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            SensorTemp = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += SENSOR_TEMP_SIZE;

            GlobalGain = metadata[currentIndex];
            currentIndex += GLOBAL_GAIN_SIZE;

            LowGain = metadata[currentIndex];
            currentIndex += LDR_GAIN_SIZE;

            HighGain = metadata[currentIndex];
            currentIndex += HDR_GAIN_SIZE;

            DummyPixelsPerRow = metadata[currentIndex];
            currentIndex += DUMMY_PIXELS_PER_ROW_SIZE;

            UnusedByteAfterDummyPixelsPerRow = metadata[currentIndex];
            currentIndex += UNUSED_BYTE_AFTER_DUMMY_PIXELS_PER_ROW_SIZE;

            ControlBlock = metadata[currentIndex];

            DeadPixelCorrection = ((metadata[currentIndex] & DEAD_PIXEL_CORRECTION_MASK) > 0);
            PixelOrderedImage = ((metadata[currentIndex] & PIXEL_ORDERED_IMAGE_MASK) > 0);
            HDRMode = ((metadata[currentIndex] & HDR_MODE) > 0);
            LowNoise = ((metadata[currentIndex] & LOW_NOISE_MASK) > 0);
            LowDarkCurrent = ((metadata[currentIndex] & LOW_DARK_CURRENT_MASK) > 0);
            HorizontalScanDirectionInvert = ((metadata[currentIndex] & HORIZONTAL_SCAN_DIRECTION_INVERT_MASK) > 0);
            VerticalScanDirectionInvert = ((metadata[currentIndex] & VERTICAL_SCAN_DIRECTION_INVERT_MASK) > 0);
            NonRowAllignedImage = ((metadata[currentIndex] & NON_ROW_ALLIGNED_IMAGE_MASK) > 0);
            currentIndex += CONTROL_BLOCK_SIZE;

            HorizontalBinning = metadata[currentIndex];
            currentIndex += HORIZONTAL_BINNING_SIZE;

            VerticalBinning = metadata[currentIndex];
            currentIndex += VERTICAL_BINNING_SIZE;

            fourBytes[0] = 0;
            Array.Copy(metadata, currentIndex, fourBytes, 1, BLACK_LEVEL_ADJUST_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            BlackLevelAdjust = BitConverter.ToInt32(fourBytes, 0);
            currentIndex += BLACK_LEVEL_ADJUST_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            BlackLevelSun = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += BLACK_LEVEL_SUN_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            twoBytes[0] = CheckForSignMagnitude(twoBytes[0], ref signed);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ShutterOpenDelay = BitConverter.ToInt16(twoBytes, 0) * signed;
            currentIndex += SHUTTER_OPEN_DELAY_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            twoBytes[0] = CheckForSignMagnitude(twoBytes[0], ref signed);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ShutterCloseDelay = BitConverter.ToInt16(twoBytes, 0) * signed;
            currentIndex += SHUTTER_CLOSE_DELAY_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            twoBytes[0] = CheckForSignMagnitude(twoBytes[0], ref signed);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            IlluminationStartDelay = BitConverter.ToInt16(twoBytes, 0) * signed;
            currentIndex += ILLUMINATION_START_DELAY_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            twoBytes[0] = CheckForSignMagnitude(twoBytes[0], ref signed);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            IlluminationStopDelay = BitConverter.ToInt16(twoBytes, 0) * signed;
            currentIndex += ILLUMINATION_STOP_DELAY_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TrackingStartRow = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TRACKING_START_ROW_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TrackingStartColumn = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TRACKING_START_COLUMN_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TrackingStopRow = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TRACKING_STOP_ROW_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TrackingStopColumn = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TRACKING_STOP_COLUMN_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ImageStartRow = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += IMAGE_START_ROW_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ImageStartColumn = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += IMAGE_START_COLUMN_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ImageStopRow = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += IMAGE_STOP_ROW_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ImageStopColumn = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += IMAGE_STOP_COLUMN_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TrackingFramesPerImageFrame = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TRACKING_FRAMES_PER_IMAGE_FRAME_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            FrameNumber = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += FRAME_NUMBER_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            StartingExposureRow = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += STARTING_EXPOSURE_ROW_SIZE;

            fourBytes[0] = 0;
            Array.Copy(metadata, currentIndex, fourBytes, 1, PRE_DUMMY_ROWS_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            PreDummyRows = (((fourBytes[1] & 0x0F) << 8) | fourBytes[0]);
            PostDummyRows = ((fourBytes[2] << 4) | ((fourBytes[1] & 0xF0) >> 4));
            currentIndex += PRE_DUMMY_ROWS_SIZE;

            NumberOfDataChannels = metadata[currentIndex];
            currentIndex += NUMBER_OF_CHANNELS_SIZE;

            ControlBlock2 = metadata[currentIndex];

            BackSideIlluminated = ((metadata[currentIndex] & BACK_SIDE_ILLUMINATED_MASK) > 0);
            GPSLocked = false;       // GPS is not valid for Version 1
            currentIndex += CONTROL_BLOCK_2_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            BaseTemperature = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += BASE_TEMPERATURE_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            FPGATemperature = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += FPGA_TEMPERATURE_SIZE;

            UnusedBytesAfterFPGATemperature = new byte[UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE];
            Array.Copy(metadata, currentIndex, UnusedBytesAfterFPGATemperature, 0, UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE);
            currentIndex += UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE;

            ControlBlockTest = metadata[currentIndex];

            UseShiftedAveraging = ((metadata[currentIndex] & TEST_USE_SHIFTED_AVERAGING_MASK) == 0);  // The firmware makes this byte 0xFF so if the 0x80 bit is 1 it means off
            IsStackedFrame = ((metadata[currentIndex] & TEST_IS_STACKED_FRAME_MASK) == 0);            // The firmware makes this byte 0xFF so if the 0x40 bit is 1 it means false
            IsHighFrame = ((metadata[currentIndex] & TEST_IS_STACKED_FRAME_MASK) == 0);               // The firmware makes this byte 0xFF so if the 0x20 bit is 1 it means false
            IsMergedFrame = ((metadata[currentIndex] & TEST_IS_STACKED_FRAME_MASK) == 0);             // The firmware makes this byte 0xFF so if the 0x10 bit is 1 it means false
            currentIndex += CONTROL_BLOCK_TEST_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, TEST_LOW_GAIN_ABSOLUTE_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TestLowGainAbsolute = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TEST_LOW_GAIN_ABSOLUTE_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, TEST_HIGH_GAIN_ABSOLUTE_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TestHighGainAbsolute = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TEST_HIGH_GAIN_ABSOLUTE_SIZE;

            Array.Copy(metadata, currentIndex, fourBytes, 0, FIRMWARE_VERSION_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            FirmwareVersion = BitConverter.ToUInt32(fourBytes, 0);
            currentIndex += FIRMWARE_VERSION_SIZE;
        }

        private void ParseMetadataVersion2(byte[] metadata, int currentIndex, DateTime captureDatetime)
        {
            FliProCameraManager fliProCameraManager = FliProCameraManager.Instance;
            byte[] twoBytes = new byte[2];
            byte[] fourBytes = new byte[4];
            byte[] eightBytes = new byte[8] { 0, 0, 0, 0, 0, 0, 0, 0 };
            int signed = 1;
            Regex regex = new Regex(@"[^a-zA-Z0-9\s\-\:\.]", (RegexOptions)0);

            CameraModel = System.Text.Encoding.UTF8.GetString(metadata, currentIndex, CAMERA_MODEL_SIZE);
            CameraModel = regex.Replace(CameraModel, "").Trim();
            currentIndex += CAMERA_MODEL_SIZE;
            CameraSerialNumber = System.Text.Encoding.UTF8.GetString(metadata, currentIndex, CAMERA_SERIAL_NUMBER_SIZE);
            CameraSerialNumber = regex.Replace(CameraSerialNumber, "").Trim();
            currentIndex += CAMERA_SERIAL_NUMBER_SIZE;

            Array.Copy(metadata, currentIndex, fourBytes, 0, FIRMWARE_VERSION_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            FirmwareVersion = BitConverter.ToUInt32(fourBytes, 0);
            currentIndex += FIRMWARE_VERSION_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TemperatureSetPoint = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += TEMPERATURE_SET_POINT_SIZE;

            CoolerDutyCycle = metadata[currentIndex];
            currentIndex += COOLER_DUTY_CYCLE_SIZE;


            //CaptureYear = (metadata[currentIndex] >> 1) + 2000;
            //CaptureMonth = ((metadata[currentIndex] & 1) << 3) + (metadata[currentIndex + 1] >> 5);
            //CaptureDay = (metadata[currentIndex + 1] & 0x1F);
            //currentIndex += (CAPTURE_YEAR_SIZE + CAPTURE_MONTH_SIZE);

            //eightBytes[0] = 0;
            //eightBytes[1] = 0;
            //eightBytes[2] = 0;
            //Array.Copy(metadata, currentIndex, eightBytes, 3, CAPTURE_TIME_SIZE);
            //if (BitConverter.IsLittleEndian)
            //{
            //    Array.Reverse(eightBytes);
            //}
            //CaptureTime = BitConverter.ToInt64(eightBytes, 0);
            //currentIndex += CAPTURE_TIME_SIZE;

            SensorPixelBitDepth = metadata[currentIndex];
            currentIndex += SENSOR_PIXEL_BIT_DEPTH_SIZE;

            DataPixelBitDepth = (metadata[currentIndex] & 0x0F) + 1;   // NOTE: DataPixelBitDepth is stored as -1
            DataZeroPoint = (metadata[currentIndex] & 0xF0) >> 4;
            currentIndex += DATA_PIXEL_BIT_DEPTH_ZERO_POINT_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            HorizontalPixels = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += HORIZONTAL_PIXELS_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            VerticalPixels = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += VERTICAL_PIXELS_SIZE;

            Array.Copy(metadata, currentIndex, fourBytes, 0, 4);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            ExposureTime = BitConverter.ToInt32(fourBytes, 0);
            currentIndex += EXPOSURE_TIME_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            SensorColdTemp = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += SENSOR_BASE_TEMP_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            SensorTemp = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += SENSOR_TEMP_SIZE;

            GlobalGain = metadata[currentIndex];
            currentIndex += GLOBAL_GAIN_SIZE;

            LowGain = metadata[currentIndex];
            currentIndex += LDR_GAIN_SIZE;

            HighGain = metadata[currentIndex];
            currentIndex += HDR_GAIN_SIZE;

            DummyPixelsPerRow = metadata[currentIndex];
            currentIndex += DUMMY_PIXELS_PER_ROW_SIZE;

            UnusedByteAfterDummyPixelsPerRow = metadata[currentIndex];
            currentIndex += UNUSED_BYTE_AFTER_DUMMY_PIXELS_PER_ROW_SIZE;

            ControlBlock = metadata[currentIndex];

            DeadPixelCorrection = ((metadata[currentIndex] & DEAD_PIXEL_CORRECTION_MASK) > 0);
            PixelOrderedImage = ((metadata[currentIndex] & PIXEL_ORDERED_IMAGE_MASK) > 0);
            HDRMode = ((metadata[currentIndex] & HDR_MODE) > 0);
            LowNoise = ((metadata[currentIndex] & LOW_NOISE_MASK) > 0);
            LowDarkCurrent = ((metadata[currentIndex] & LOW_DARK_CURRENT_MASK) > 0);
            HorizontalScanDirectionInvert = ((metadata[currentIndex] & HORIZONTAL_SCAN_DIRECTION_INVERT_MASK) > 0);
            VerticalScanDirectionInvert = ((metadata[currentIndex] & VERTICAL_SCAN_DIRECTION_INVERT_MASK) > 0);
            NonRowAllignedImage = ((metadata[currentIndex] & NON_ROW_ALLIGNED_IMAGE_MASK) > 0);
            currentIndex += CONTROL_BLOCK_SIZE;

            HorizontalBinning = metadata[currentIndex];
            currentIndex += HORIZONTAL_BINNING_SIZE;

            VerticalBinning = metadata[currentIndex];
            currentIndex += VERTICAL_BINNING_SIZE;

            fourBytes[0] = 0;
            Array.Copy(metadata, currentIndex, fourBytes, 1, BLACK_LEVEL_ADJUST_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            BlackLevelAdjust = BitConverter.ToInt32(fourBytes, 0);
            currentIndex += BLACK_LEVEL_ADJUST_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            BlackLevelSun = BitConverter.ToInt16(twoBytes, 0);
            currentIndex += BLACK_LEVEL_SUN_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            twoBytes[0] = CheckForSignMagnitude(twoBytes[0], ref signed);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ShutterOpenDelay = BitConverter.ToInt16(twoBytes, 0) * signed;
            currentIndex += SHUTTER_OPEN_DELAY_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            twoBytes[0] = CheckForSignMagnitude(twoBytes[0], ref signed);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ShutterCloseDelay = BitConverter.ToInt16(twoBytes, 0) * signed;
            currentIndex += SHUTTER_CLOSE_DELAY_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            twoBytes[0] = CheckForSignMagnitude(twoBytes[0], ref signed);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            IlluminationStartDelay = BitConverter.ToInt16(twoBytes, 0) * signed;
            currentIndex += ILLUMINATION_START_DELAY_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            twoBytes[0] = CheckForSignMagnitude(twoBytes[0], ref signed);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            IlluminationStopDelay = BitConverter.ToInt16(twoBytes, 0) * signed;
            currentIndex += ILLUMINATION_STOP_DELAY_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TrackingStartRow = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TRACKING_START_ROW_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TrackingStartColumn = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TRACKING_START_COLUMN_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TrackingStopRow = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TRACKING_STOP_ROW_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TrackingStopColumn = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TRACKING_STOP_COLUMN_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ImageStartRow = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += IMAGE_START_ROW_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ImageStartColumn = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += IMAGE_START_COLUMN_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ImageStopRow = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += IMAGE_STOP_ROW_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            ImageStopColumn = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += IMAGE_STOP_COLUMN_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TrackingFramesPerImageFrame = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TRACKING_FRAMES_PER_IMAGE_FRAME_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            FrameNumber = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += FRAME_NUMBER_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            StartingExposureRow = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += STARTING_EXPOSURE_ROW_SIZE;

            fourBytes[0] = 0;
            Array.Copy(metadata, currentIndex, fourBytes, 1, PRE_DUMMY_ROWS_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            PreDummyRows = (((fourBytes[1] & 0x0F) << 8) | fourBytes[0]);
            PostDummyRows = ((fourBytes[2] << 4) | ((fourBytes[1] & 0xF0) >> 4));
            currentIndex += PRE_DUMMY_ROWS_SIZE;

            NumberOfDataChannels = metadata[currentIndex];
            currentIndex += NUMBER_OF_CHANNELS_SIZE;

            ControlBlock2 = metadata[currentIndex];

            BackSideIlluminated = ((metadata[currentIndex] & BACK_SIDE_ILLUMINATED_MASK) > 0);
            GPSLocked = ((metadata[currentIndex] & GPS_LOCKED_MASK) > 0);
            currentIndex += CONTROL_BLOCK_2_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            BaseTemperature = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += BASE_TEMPERATURE_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, 2);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            FPGATemperature = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += FPGA_TEMPERATURE_SIZE;

            UnusedBytesAfterFPGATemperature = new byte[UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE];
            Array.Copy(metadata, currentIndex, UnusedBytesAfterFPGATemperature, 0, UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE);
            currentIndex += UNUSED_BYTES_AFTER_FPGA_TEMPERATURE_SIZE;

            ControlBlockTest = metadata[currentIndex];

            UseShiftedAveraging = ((metadata[currentIndex] & TEST_USE_SHIFTED_AVERAGING_MASK) == 0);  // The firmware makes this byte 0xFF so if the 0x80 bit is 1 it means off
            IsStackedFrame = ((metadata[currentIndex] & TEST_IS_STACKED_FRAME_MASK) == 0);            // The firmware makes this byte 0xFF so if the 0x40 bit is 1 it means false
            IsHighFrame = ((metadata[currentIndex] & TEST_IS_STACKED_FRAME_MASK) == 0);               // The firmware makes this byte 0xFF so if the 0x20 bit is 1 it means false
            IsMergedFrame = ((metadata[currentIndex] & TEST_IS_STACKED_FRAME_MASK) == 0);             // The firmware makes this byte 0xFF so if the 0x10 bit is 1 it means false
            currentIndex += CONTROL_BLOCK_TEST_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, TEST_LOW_GAIN_ABSOLUTE_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TestLowGainAbsolute = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TEST_LOW_GAIN_ABSOLUTE_SIZE;

            Array.Copy(metadata, currentIndex, twoBytes, 0, TEST_HIGH_GAIN_ABSOLUTE_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(twoBytes);
            }
            TestHighGainAbsolute = BitConverter.ToUInt16(twoBytes, 0);
            currentIndex += TEST_HIGH_GAIN_ABSOLUTE_SIZE;

            CaptureString = Encoding.UTF8.GetString(metadata, currentIndex, TIMESTAMP_SIZE);
            CaptureString = regex.Replace(CaptureString, "").Trim();
            bool replacedTime = ParseTimestamp(ref CaptureString, captureDatetime);
            Array.Copy(Encoding.ASCII.GetBytes(CaptureString), 0, metadata, currentIndex, CaptureString.Length);
            if (replacedTime)
            {
                GPSLocked = false;
            }
            currentIndex += TIMESTAMP_SIZE;

            Array.Copy(metadata, currentIndex, fourBytes, 0, LATITUDE_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            LatitudeRaw = BitConverter.ToUInt32(fourBytes, 0);
            ParseDirection(LatitudeRaw, ref LatitudeNorth, ref LatitudeDegrees, ref LatitudeMinutes);
            currentIndex += LATITUDE_SIZE;

            Array.Copy(metadata, currentIndex, fourBytes, 0, LONGITUDE_SIZE);
            if (BitConverter.IsLittleEndian)
            {
                Array.Reverse(fourBytes);
            }
            LongitudeRaw = BitConverter.ToUInt32(fourBytes, 0);
            ParseDirection(LongitudeRaw, ref LongitudeEast, ref LongitudeDegrees, ref LongitudeMinutes);
            currentIndex += LONGITUDE_SIZE;

            // We need to see if this is comming from a file so that we don't overwrite it with the newer current version
            FliPilotVersion = System.Text.Encoding.UTF8.GetString(metadata, currentIndex, FLIPILOT_VERSION_SIZE);
            FliPilotVersion = regex.Replace(FliPilotVersion, "").Trim();
            if (String.IsNullOrEmpty(FliPilotVersion))
            {
                FliPilotVersion = Assembly.GetExecutingAssembly().GetName().Version.ToString();
                FliPilotVersion = FliPilotVersion.Substring(0, Math.Min(FliPilotVersion.Length, FLIPILOT_VERSION_SIZE));   // Make sure its not too long
                Array.Copy(Encoding.ASCII.GetBytes(FliPilotVersion), 0, metadata, currentIndex, FliPilotVersion.Length);
            }
            currentIndex += FLIPILOT_VERSION_SIZE;

            // We need to see if this is comming from a file so that we don't overwrite it with the newer current version
            LibFliProVersion = System.Text.Encoding.UTF8.GetString(metadata, currentIndex, LIB_FLI_PRO_VERSION_SIZE);
            LibFliProVersion = regex.Replace(LibFliProVersion, "").Trim();
            if (String.IsNullOrEmpty(LibFliProVersion))
            {
                LibFliProVersion = fliProCameraManager.GetAPIVersion();
                LibFliProVersion = LibFliProVersion.Substring(0, Math.Min(LibFliProVersion.Length, LIB_FLI_PRO_VERSION_SIZE));   // Make sure its not too long
                Array.Copy(Encoding.ASCII.GetBytes(LibFliProVersion), 0, metadata, currentIndex, LibFliProVersion.Length);
            }
            currentIndex += LIB_FLI_PRO_VERSION_SIZE;

            HorizontalPixelSize = metadata[currentIndex];
            currentIndex += HORIZONTAL_PIXEL_SIZE;
            VerticalPixelSize = metadata[currentIndex];
            currentIndex += VERTICAL_PIXEL_SIZE;

            UnusedBytesAfterLongitude = new byte[UNUSED_BYTES_AFTER_LONGITUDE_SIZE];
            Array.Copy(metadata, currentIndex, UnusedBytesAfterLongitude, 0, UNUSED_BYTES_AFTER_LONGITUDE_SIZE);
            currentIndex += UNUSED_BYTES_AFTER_LONGITUDE_SIZE;

        }

        public void ParseDirection(uint reading, ref bool directionFlag, ref int degrees, ref double minutes)
        {
            directionFlag = ((reading & DIRECTION_FLAG_MASK) == DIRECTION_FLAG_MASK);
            degrees = (int)((reading & DIRECTION_DEGREES_MASK) / DIRECTION_DEGREES_DIVISOR);
            minutes = (((double)(reading & DIRECTION_DEGREES_MASK) / DIRECTION_DEGREES_DIVISOR) - degrees) * 60;
        }

        public void ParseDirection(double location, ref uint rawData, ref bool directionFlag, ref int degrees, ref double minutes)
        {
            if (location < 0)
            {
                // This is a southern or western reading
                rawData = Convert.ToUInt32(location * -1 * DIRECTION_DEGREES_DIVISOR);
            }
            else
            {
                rawData = Convert.ToUInt32(location * DIRECTION_DEGREES_DIVISOR) + DIRECTION_FLAG_MASK;
            }
            directionFlag = ((rawData & DIRECTION_FLAG_MASK) == DIRECTION_FLAG_MASK);
            degrees = (int)((rawData & DIRECTION_DEGREES_MASK) / DIRECTION_DEGREES_DIVISOR);
            minutes = (((double)(rawData & DIRECTION_DEGREES_MASK) / DIRECTION_DEGREES_DIVISOR) - degrees) * 60;
        }

        public bool ParseTimestamp(ref string timeString, DateTime captureDatetime)
        {
            bool replacedTime = true;

            if (timeString.Length < 26)
            {
                CaptureYear = captureDatetime.Year;
                CaptureMonth = captureDatetime.Month;
                CaptureDay = captureDatetime.Day;
                CaptureHour = captureDatetime.Hour;
                CaptureMinute = captureDatetime.Minute;
                CaptureSecond = captureDatetime.Second;
                CaptureNanoseconds = captureDatetime.Millisecond * 1000;
                timeString = CaptureYear.ToString("0000") + "-" + CaptureMonth.ToString("00") + "-" + CaptureDay.ToString("00") + "T" +
                    CaptureHour.ToString("00") + ":" + CaptureMinute.ToString("00") + ":" + CaptureSecond.ToString("00") + "." + CaptureNanoseconds.ToString("000000000");
            }
            else
            {
                // yyyy-mm-ddTHH:mm:ss.ffffff
                try
                {
                    CaptureYear = Convert.ToInt32(timeString.Substring(0, 4));
                    CaptureMonth = Convert.ToInt32(timeString.Substring(5, 2));
                    CaptureDay = Convert.ToInt32(timeString.Substring(8, 2));
                    CaptureHour = Convert.ToInt32(timeString.Substring(11, 2));
                    CaptureMinute = Convert.ToInt32(timeString.Substring(14, 2));
                    CaptureSecond = Convert.ToInt32(timeString.Substring(17, 2));
                    CaptureNanoseconds = Convert.ToInt32(timeString.Substring(20, 9));
                    replacedTime = false;
                }
                catch
                {
                    CaptureYear = 0;
                }
                if (CaptureYear < 2018)
                {
                    CaptureYear = captureDatetime.Year;
                    CaptureMonth = captureDatetime.Month;
                    CaptureDay = captureDatetime.Day;
                    CaptureHour = captureDatetime.Hour;
                    CaptureMinute = captureDatetime.Minute;
                    CaptureSecond = captureDatetime.Second;
                    CaptureNanoseconds = captureDatetime.Millisecond * 1000;
                    timeString = CaptureYear.ToString("0000") + "-" + CaptureMonth.ToString("00") + "-" + CaptureDay.ToString("00") + "T" +
                        CaptureHour.ToString("00") + ":" + CaptureMinute.ToString("00") + ":" + CaptureSecond.ToString("00") + "." + CaptureNanoseconds.ToString("000000000");
                }
            }

            return replacedTime;
        }

        private void StoreValue(object value, byte[] metaData, ref int arrayIndex, int size)
        {
            byte[] dataArray = null;

            switch (size)
            {
                case 2:
                    ushort targetShort = Convert.ToUInt16(value);
                    dataArray = BitConverter.GetBytes(targetShort);
                    break;
                case 4:
                    uint targetInt = Convert.ToUInt32(value);
                    dataArray = BitConverter.GetBytes(targetInt);
                    break;
                case 8:
                    ulong targetLong = Convert.ToUInt64(value);
                    dataArray = BitConverter.GetBytes(targetLong);
                    break;
            }

            if (dataArray != null)
            {
                Array.Reverse(dataArray);
                Array.Copy(dataArray, 0, metaData, arrayIndex, size);
                arrayIndex += size;
            }
        }

        #endregion Private Functions

    }
}
