﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// Application and System Information class
//
// Copyright(c) 2014-2016 AppliedLogix, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//
// Renamed to FliUtils on 10/24/2017 to hide AppliedLogix name from customers
//
//**************************************************************

using Microsoft.Win32;
using System;
using System.DirectoryServices.AccountManagement;
using System.IO;
using System.Reflection;

namespace FliUtils.App
{
    /// <summary>
    /// Singleton class providing information about the
    /// current application and its environment
    /// </summary>
    public class AppInfo
    {
        private static volatile AppInfo _instance;      // Singleton class instance
        private static object syncRoot = new Object();  // mutex for creating instance

        // Application assembly information
        public string AppName { get; private set; }
        public string AppVersion { get; private set; }
        public string AppDescription { get; private set; }
        public string AppCompany { get; private set; }
        public string AppCopyright { get; private set; }
        public string AppArchitecture { get; private set; }
        public string AppPath { get; private set; }
        public string AppDir { get; private set; } 
        public string UserName { get; private set; }
        public string UserFullName { get; private set; }
        public string SystemName { get; private set; }
        public string SystemArchitecture { get; private set; }
        public string OSName { get; private set; }
        public string OSServicePack { get; private set; }
        public string OSBuild { get; private set; }

        // Backing store directory paths
        public string AppBaseDir { get; private set; }
        //public string AppConfigDir { get; private set; }

        // User default directory Paths
        public string UserBaseDir { get; private set; }
        //public string UserDataDir { get; private set; }

        // Temporary file directory path
        public string TempDir { get; private set; }

        // True if bacing store directories exist
        public bool AppDirectoriesOK { get; private set; }
        public bool UserDirectoriesOK { get; private set; }

        /// <summary>
        /// Private constructor - must use Instance to get reference to singleton
        /// </summary>
        private AppInfo()
        {
            var appAssembly = Assembly.GetEntryAssembly();
            var assemblyName = appAssembly.GetName();
            AppName = assemblyName.Name;
            AppPath = appAssembly.Location;
            AppDir = Path.GetDirectoryName(AppPath);
            AppVersion = assemblyName.Version.ToString();
            AppArchitecture = assemblyName.ProcessorArchitecture.ToString();

            // Additional Assembly attributes
            foreach (Attribute attr in Attribute.GetCustomAttributes(appAssembly))
            {
                var type = attr.GetType();
                if (type == typeof(AssemblyCompanyAttribute))
                {
                    AppCompany = ((AssemblyCompanyAttribute)attr).Company;
                }
                else if (type == typeof(AssemblyCopyrightAttribute))
                {
                    AppCopyright = ((AssemblyCopyrightAttribute)attr).Copyright;
                }
                else if (type == typeof(AssemblyDescriptionAttribute))
                {
                    AppDescription = ((AssemblyDescriptionAttribute)attr).Description;
                }
            }

            string dataDir = "\\" + AppCompany;
            AppBaseDir = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData) + dataDir;
            //AppConfigDir = AppBaseDir + "\\Config";

            UserBaseDir = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + dataDir;
            //UserDataDir = UserBaseDir + "\\data";

            TempDir = System.IO.Path.GetTempPath() + "\\" + AppCompany;

            SystemName = Environment.MachineName;
            SystemArchitecture = GetSystemArchitecture();
            OSName = GetOSName();
            OSServicePack = GetOSServicePack();
            OSBuild = GetOSBuild();

            UserName = Environment.UserName;
            UserFullName = GetUserFullName(UserName);

            // Create directories if needed
            AppDirectoriesOK = true;
            try
            {
                Directory.CreateDirectory(AppBaseDir);
                //Directory.CreateDirectory(AppConfigDir);
            }
            catch (Exception)
            {
                AppDirectoriesOK = false;
            }

            UserDirectoriesOK = true;
            try
            {
                Directory.CreateDirectory(UserBaseDir);
                //Directory.CreateDirectory(UserDataDir);
                DeleteTempDir();
                Directory.CreateDirectory(TempDir);
            }
            catch (Exception)
            {
                UserDirectoriesOK = false;
            }
        }

        ~AppInfo()
        {
            DeleteTempDir();
        }

        //-------------------------------------------
        /// <summary>
        /// Returns a reference to the single instance of this class.
        /// </summary>
        public static AppInfo Instance
        {
            get
            {
                if (_instance == null)
                {
                    lock (syncRoot)
                    {
                        if (_instance == null)
                        {
                            _instance = new AppInfo();
                        }
                    }
                }
                return _instance;
            }
        }

        /// <summary>
        /// Backing Store base directory for the specified device
        /// </summary>
        public string GetAppDeviceBaseDir(string deviceID)
        {
            string dir = AppBaseDir + "\\" + deviceID;
            Directory.CreateDirectory(dir);
            return dir;
        }

        /// <summary>
        /// Backing Store data directory for the specified device
        /// </summary>
        // public string GetAppToolDataDir(string deviceID)
        // {
            // string dir = GetAppToolBaseDir(toolID) + "\\data";
            // Directory.CreateDirectory(dir);
            // return dir;
        // }

        /// <summary>
        /// Temporary directory for the specified device.
        /// Automatically deleted on program exit
        /// </summary>
        public string GetAppDeviceTempDir(string deviceID)
        {
           string dir = TempDir + "\\" + deviceID;
           Directory.CreateDirectory(dir);
            return dir;
        }

        /// <summary>
        /// User's data directory for the specified device
        /// </summary>
        //public string GetUserDeviceDataDir(string deviceID)
        //{
        //    string dir = UserDataDir + "\\" + deviceID;
        //    Directory.CreateDirectory(dir);
        //    return dir;
        //}

        /// <summary>
        /// Safely try to delete application's temporary directory if it exists
        /// Take care not to inadvertently delete useful data
        /// </summary>
        public void DeleteTempDir()
        {
            try
            {
                if (TempDir.StartsWith(System.IO.Path.GetTempPath()) &&
                    TempDir.Contains(AppCompany))
                {
                    Directory.Delete(TempDir, true);
                }
            }
            catch (Exception)
            {
            }
        }

        private string GetOSName()
        {
            string osName = "";
            var key = Registry.LocalMachine.OpenSubKey(@"Software\Microsoft\Windows NT\CurrentVersion");
            if (key != null)
            {
                object osObj = key.GetValue("ProductName");
                if (osObj != null)
                {
                    osName = osObj as string;
                }
            }
            if ((osName == null) || (osName == ""))
            {
                osName = "unknown";
            }
            return osName;
        }

        private string GetOSServicePack()
        {
            string osSP = "";
            var key = Registry.LocalMachine.OpenSubKey(@"Software\Microsoft\Windows NT\CurrentVersion");
            if (key != null)
            {
                object osObj = key.GetValue("CSDVersion");
                if (osObj != null)
                {
                    osSP = osObj as string;
                }
            }
            if ((osSP == null))
            {
                osSP = "";
            }
            return osSP;
        }

        private string GetOSBuild()
        {
            string osBuild = "";
            var key = Registry.LocalMachine.OpenSubKey(@"Software\Microsoft\Windows NT\CurrentVersion");
            if (key != null)
            {
                osBuild = key.GetValue("CurrentBuild") as string;
                if ((osBuild == null))
                {
                    osBuild = "";
                }
            }
            return osBuild;
        }

        private string GetSystemArchitecture()
        {
            string arch = "";
            var key = Registry.LocalMachine.OpenSubKey(@"System\CurrentControlSet\Control\Session Manager\Environment");
            if (key != null)
            {
                arch = key.GetValue("PROCESSOR_ARCHITECTURE") as string;
            }
            if ((arch == null) || arch == "")
            {
                arch = "unknown";
            }
            return arch;
        }

        private string GetUserFullName(string defaultname)
        {
            string fullName = UserPrincipal.Current.DisplayName;
            if ((fullName == null) || (fullName == ""))
            {
                fullName = defaultname;
            }
            return fullName;
        }
    }
}
