﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// Base class for supporting observable properties
//
// Copyright(c) 2014-2016 AppliedLogix, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//**************************************************************
using System.ComponentModel;
using System.Runtime.CompilerServices;

namespace FliUtils.Mvvm
{
    /// <summary>
    /// Simple base class for classes supporting observable properties
    /// </summary>
    public class ObservableProperties : INotifyPropertyChanged
    {
        //**************************************************************
        #region Properties

        /// <summary>
        /// Force SetValue to change property if true
        /// Note: for initialization only
        /// </summary>
        protected bool ForcePropertyChange { get; set; }

        #endregion Properties

        //**************************************************************
        #region Methods

        /// <summary>
        /// Assigns a new value to a property and raises the
        /// PropertyChanged event if needed
        /// </summary>
        /// <typeparam name="t">property type</typeparam>
        /// <param name="currentvalue">reference to property backing variable of type t</param>
        /// <param name="newvalue">new value of type t</param>
        /// <param name="name">Property name (defaults to the calling property name)</param>
        /// <returns>true if value was changed</returns>
        protected bool SetValue<t>(ref t currentvalue, t newvalue, [CallerMemberName] string name = null)
        {
            bool changed = false;
            if ((currentvalue == null) || !(currentvalue.Equals(newvalue)) || ForcePropertyChange)
            {
                currentvalue = newvalue;
                OnPropertyChanged(name);
                changed = true;
            }
            return changed;
        }

        #endregion Methods

        //**************************************************************
        #region INotifyPropertyChanged
        
        /// <summary>
        /// PropertyChanged event - raised when an observable property
        /// value has changed.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// Called to raise PropertyChanged event for the
        /// property specified by name, or for all properties if 
        /// explicitly set to null.
        /// </summary>
        /// <param name="name">Property name (defaults to the calling property name)</param>
        protected void OnPropertyChanged([CallerMemberName] string name = null)
        {
            PropertyChangedEventHandler handler = PropertyChanged;
            if (handler != null)
            {
                handler(this, new PropertyChangedEventArgs(name));
            }
        }

        /// <summary>
        /// Called to raise PropertyChanged event for the
        /// property specified by name, or for all properties if 
        /// explicitly set to null.
        /// Needed the sender to be specifically different from "this".
        /// </summary>
        /// <param name="sender">The specific control which can be other than simply "this"</param>
        /// <param name="name">Property name (defaults to the calling property name)</param>
        protected void OnPropertyChanged(object sender, [CallerMemberName] string name = null)
        {
            PropertyChangedEventHandler handler = PropertyChanged;
            if (handler != null)
            {
                handler(sender, new PropertyChangedEventArgs(name));
            }
        }

        #endregion // INotifyPropertyChanged
    }
}
