﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// USB Device Connection/Disconnection Event and Information Class
//
// Copyright(c) 2016 AppliedLogix, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//**************************************************************
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Management;
using System.Text;
using System.Threading.Tasks;

namespace FliUtils.Usb
{
    /// <summary>
    /// Class to raise events on USB device connection/disconnection
    /// and provide basic device information
    /// </summary>
    public class USBMonitor
    {
        //**************************************************************
        #region Constructor

        public USBMonitor()
        {
            ManagementScope scope = new ManagementScope("root\\CIMV2");
            WqlEventQuery usbQuery = new WqlEventQuery();
            usbQuery.EventClassName = "__InstanceOperationEvent";
            usbQuery.Condition = "TargetInstance ISA 'Win32_USBControllerDevice'";
            usbQuery.WithinInterval = new TimeSpan(0, 0, 1);
            usbWatcher = new ManagementEventWatcher(scope, usbQuery);
            usbWatcher.EventArrived += USBWMIEventHandler;
        }

        #endregion Constructor

        //**************************************************************
        #region Private Properties

        private ManagementEventWatcher usbWatcher;

        #endregion Private Properties

        //**************************************************************
        #region Public Properties

        public event USBMonitorEventHandler Connected;
        public event USBMonitorEventHandler Disconnected;

        /// <summary>
        /// Collection of DeviceID definitions for which connection/disconnection
        /// events will be raised
        /// </summary>
        public Collection<USBDeviceID> DeviceIDs
        {
            get { return _DeviceIDs; }
        }
        private Collection<USBDeviceID> _DeviceIDs = new Collection<USBDeviceID>();

        #endregion Public Properties

        //**************************************************************
        #region Public Methods

        /// <summary>
        /// Rescan for USB device connections and notify of any devices already connected
        /// </summary>
        public void Rescan()
        {
            // Raise "Connected" events for any already connected devices
            var devicelist = GetDevices();
            foreach (var devinfo in devicelist)
            {
                Connected(this, new USBMonitorEventArgs(devinfo));
            }
        }

        /// <summary>
        /// Start monitoring for USB device (dis)connections
        /// optionally notifying of any devices already connected
        /// </summary>
        /// <param name="includeExisting">also notify for existing devices if true (default)</param>
        public void Start(bool includeExisting = true)
        {
            if (includeExisting)
            {
                // Raise "Connected" events for any already connected devices
                var devicelist = GetDevices();
                foreach(var devinfo in devicelist)
                {
                    Connected(this, new USBMonitorEventArgs(devinfo));
                }
            }
            // Start WMI watcher for new connections/disconnections
            usbWatcher.Start();
        }

        /// <summary>
        /// Stop monitoring for USB device 9dis)connections
        /// </summary>
        public void Stop()
        {
            usbWatcher.Stop();
        }

        /// <summary>
        /// Get list of all currently connected USB devices that match the specified DeviceID(s)
        /// </summary>
        /// <returns></returns>
        List<USBDeviceInfo> GetDevices()
        {
            var devices = new List<USBDeviceInfo>();

            ManagementObjectCollection devCollection;
            string query = @"Select * From Win32_USBControllerDevice";
            using (var searcher = new ManagementObjectSearcher(query))
            {
                devCollection = searcher.Get();
            }
            foreach (var device in devCollection)
            {
                var path = device.Properties["Dependent"].Value as string;
                var info = new USBDeviceInfo(path, true, DeviceIDs);
                if(info.Valid)
                {
                    devices.Add(info);
                }
            }
            return devices;
        }

        #endregion Public Methods

        //**************************************************************
        #region Private Methods
        
        /// <summary>
        /// Event handler for WMI USB monitor
        /// </summary>
        private void USBWMIEventHandler(object sender, EventArrivedEventArgs e)
        {
            var eventtype = e.NewEvent.ClassPath.ClassName;
            var mbo = e.NewEvent.Properties["TargetInstance"].Value as ManagementBaseObject;
            var path = mbo.Properties["Dependent"].Value as string;
            bool connect = (eventtype == "__InstanceCreationEvent");
            bool disconnect = (eventtype == "__InstanceDeletionEvent");

            var info = new USBDeviceInfo(path, connect, DeviceIDs);
            if(info.Valid)
            {
                if(connect)
                {
                    Connected(this, new USBMonitorEventArgs(info));
                }
                else if(disconnect)
                {
                    Disconnected(this, new USBMonitorEventArgs(info));
                }
            }
        }

        /// <summary>
        /// Handler for connection events
        /// </summary>
        private void OnConnect(USBMonitorEventArgs e)
        {
            if (Connected != null)
            {
                Connected(this, e);
            }
        }

        /// <summary>
        /// Handler for disconnection events
        /// </summary>
        private void OnDisconnect(USBMonitorEventArgs e)
        {
            if (Disconnected != null)
            {
                Disconnected(this, e);
            }
        }

        #endregion Private Regions

    }

    /// <summary>
    /// USB Monitor Event Handler Delegate
    /// </summary>
    public delegate void USBMonitorEventHandler(object sender, USBMonitorEventArgs e);

    /// <summary>
    /// USB Monitor Event Arguments Class
    /// </summary>
    public class USBMonitorEventArgs : EventArgs
    {
        public USBDeviceInfo Info { get; private set; }

        public USBMonitorEventArgs(string path = "") : this(new USBDeviceInfo(path))
        {
        }

        public USBMonitorEventArgs(USBDeviceInfo info)
        {
            Info = info;
        }
    }
}

